"""Provide functionality for ghostwriting test modules."""
import argparse
import enum
import inspect
import pathlib
import re
import sys
import textwrap
from typing import Optional, Tuple, List, overload, Union, Callable, Any, cast

import hypothesis.strategies
import hypothesis.strategies._internal.collections
import hypothesis.strategies._internal.lazy
import icontract._checkers
import icontract._represent

import icontract_hypothesis
import icontract_hypothesis.pyicontract_hypothesis._general as _general


@icontract.invariant(lambda self: (self.module_name is None) ^ (self.path is None))
class Params:
    """Represent parameters of the command "ghostwrite"."""

    @icontract.require(lambda module_name, path: (module_name is None) ^ (path is None))
    def __init__(
        self,
        module_name: Optional[str],
        path: Optional[pathlib.Path],
        output: Optional[pathlib.Path],
        explicit: bool,
        bare: bool,
    ) -> None:
        """Initialize with the given values."""
        self.module_name = module_name
        self.path = path
        self.output = output
        self.explicit = explicit
        self.bare = bare


def parse_params(args: argparse.Namespace) -> Tuple[Optional[Params], List[str]]:
    """
    Try to parse the parameters of the command "ghostwrite".

    Return (parsed parameters, errors if any).
    """
    output = pathlib.Path(args.output) if args.output != "-" else None

    path = None  # type: Optional[pathlib.Path]

    if args.path is not None:
        try:
            path = pathlib.Path(args.path)
        except Exception as err:
            return None, [f"Failed to parse --path {args.path!r}: {err}"]

    return (
        Params(
            module_name=args.module,
            path=path,
            output=output,
            explicit=args.explicit,
            bare=args.bare,
        ),
        [],
    )


@overload
def _indent_but_first(
    lines: List[str], level: int = 1
) -> List[str]:  # pylint: disable=all
    ...


@overload
def _indent_but_first(lines: str, level: int = 1) -> str:  # pylint: disable=all
    ...


def _indent_but_first(
    lines: Union[List[str], str], level: int = 1
) -> Union[str, List[str]]:
    r"""
    Indents the text by 4 spaces.

    >>> _indent_but_first([''], 0)
    ['']

    >>> _indent_but_first(['test'], 1)
    ['test']

    >>> _indent_but_first(['test', '', 'me'], 1)
    ['test', '', '    me']

    >>> _indent_but_first('test\n\nme', 1)
    'test\n\n    me'
    """
    if isinstance(lines, str):
        result = lines.splitlines()
        for i in range(1, len(result)):
            if len(result[i]) > 0:
                result[i] = "    " * level + result[i]

        return "\n".join(result)

    elif isinstance(lines, list):
        result = lines[:]
        for i in range(1, len(result)):
            if len(result[i]) > 0:
                result[i] = "    " * level + result[i]

        return result

    else:
        raise AssertionError("Unhandled input: {}".format(lines))


def _ghostwrite_condition_code(condition: Callable[..., Any]) -> str:
    """Ghostwrite the code representing the condition in an assumption."""
    if not icontract._represent.is_lambda(a_function=condition):
        sign = inspect.signature(condition)
        args = ", ".join(param for param in sign.parameters.keys())

        return "{}({})".format(condition.__name__, args)

    # We need to extract the source code corresponding to the decorator since
    # inspect.getsource() is broken with lambdas.

    # Find the line corresponding to the condition lambda
    lines, condition_lineno = inspect.findsource(condition)
    filename = inspect.getsourcefile(condition)
    assert filename is not None

    decorator_inspection = icontract._represent.inspect_decorator(
        lines=lines, lineno=condition_lineno, filename=filename
    )

    lambda_inspection = icontract._represent.find_lambda_condition(
        decorator_inspection=decorator_inspection
    )

    assert (
        lambda_inspection is not None
    ), "Expected lambda_inspection to be non-None if is_lambda is True on: {}".format(
        condition
    )

    return lambda_inspection.text


@icontract.ensure(lambda result: len(result[1]) > 0 or not result[0].endswith("\n"))
def _ghostwrite_test_function(
    module_name: str, point: _general.FunctionPoint, explicit: bool
) -> Tuple[str, List[str]]:
    """
    Ghostwrite a test function for the given function point.

    The result needs to be properly indented afterwards.

    Return (code, errors if any)
    """
    test_func = ""

    if not explicit:
        test_func = textwrap.dedent(
            f"""\
            def test_{point.func.__name__}(self) -> None:
                icontract_hypothesis.test_with_inferred_strategy(
                    {module_name}.{point.func.__name__})
            """
        ).strip()
        return test_func, []

    try:
        strategy = icontract_hypothesis.infer_strategy(func=point.func)
    except Exception as err:
        return "", [
            f"Failed to infer a search strategy for the function "
            f"on line {point.first_row}: {point.func.__name__}. "
            f"The exception was:\n{err}"
        ]

    # If there is no filtering on multiple arguments, unpack the argument strategies and
    # re-pack them into a hypothesis.given decorator with argument names as keyword arguments
    if (
        isinstance(strategy, hypothesis.strategies._internal.lazy.LazyStrategy)
        and isinstance(
            strategy.wrapped_strategy,
            hypothesis.strategies._internal.collections.FixedKeysDictStrategy,
        )
        and len(strategy.wrapped_strategy.keys) > 0
    ):
        fixed_dict_st = strategy.wrapped_strategy

        assert isinstance(
            fixed_dict_st.mapped_strategy,
            hypothesis.strategies._internal.collections.TupleStrategy,
        )

        given_args_lines = []  # type: List[str]

        for arg_name, arg_strategy in zip(
            fixed_dict_st.keys, fixed_dict_st.mapped_strategy.element_strategies
        ):
            line = f"{arg_name}={arg_strategy}"
            assert not line.endswith("\n")
            given_args_lines.append(line)

        given = textwrap.dedent(
            """\
            @given(
                {}
            )"""
        ).format(_indent_but_first(",\n".join(given_args_lines), level=1))
    else:
        given = textwrap.dedent(
            f"""\
            @given(
                {strategy}
            )"""
        )

    assert not re.match(
        r"^\s", given
    ), f"Unexpected whitespace at the begining of given: {given!r}"
    assert not given.endswith("\n")

    test_func = textwrap.dedent(
        """\
        def test_{func}(self) -> None:
            {given}
            def execute(kwargs) -> None:
                {module}.{func}(**kwargs)

            execute()"""
    ).format(
        module=module_name,
        func=point.func.__name__,
        given=_indent_but_first(given, level=1),
    )

    return test_func, []


def _ghostwrite_for_function_points(
    points: List[_general.FunctionPoint],
    module_name: str,
    explicit: bool,
    bare: bool,
) -> Tuple[str, List[str]]:
    """
    Ghostwrite a test case for the given function points.

    Return (generated code, errors if any).
    """
    errors = []  # type: List[str]
    test_funcs = []  # type: List[str]
    for point in points:
        test_func, test_func_errors = _ghostwrite_test_function(
            module_name=module_name, point=point, explicit=explicit
        )
        if test_func_errors:
            errors.extend(test_func_errors)
        else:
            test_funcs.append(test_func)

    if errors:
        return "", errors

    if bare:
        return "\n\n".join(test_funcs), []

    blocks = []  # type: List[str]

    header = "\n\n".join(
        [
            '"""Test {} with inferred Hypothesis strategies."""'.format(module_name),
            "import unittest",
        ]
        + (
            ["from hypothesis import given"]
            if explicit
            else ["import icontract_hypothesis"]
        )
        + ["import {}".format(module_name)]
    )
    blocks.append(header)

    if len(points) == 0:
        blocks.append(
            textwrap.dedent(
                '''\
                class TestWithInferredStrategies(unittest.TestCase):
                    """Test all functions from {0} with inferred Hypothesis strategies."""
                    # Either there are no functions in {0} or all the functions were excluded.
                '''.format(
                    module_name
                )
            ).strip()
        )
    else:
        body = "\n\n".join(test_funcs)

        test_case = [
            textwrap.dedent(
                '''\
                class TestWithInferredStrategies(unittest.TestCase):
                    """Test all functions from {module} with inferred Hypothesis strategies."""

                    {body}'''
            ).format(
                module=module_name,
                body="\n".join(_indent_but_first(lines=body.splitlines(), level=1)),
            )
        ]
        blocks.append("".join(test_case))

    blocks.append(
        textwrap.dedent(
            """\
        if __name__ == '__main__':
            unittest.main()
        """
        )
    )

    return "\n\n\n".join(blocks), []


_IDENTIFIER_RE = re.compile(r"^[a-zA-Z_][a-zA-Z_0-9]*$")


def _qualified_module_name_from_path(path: pathlib.Path, sys_path: List[str]) -> str:
    """
    Infer the qualified (dotted) module name from its path.

    If the path does not point to a proper Python module, the module name is computed
    as the stem of the ``path`` with all the non-identifier characters replaced by an
    underscore ("_").

    :param path: to the module
    :param sys_path: the value of ``sys.path``
    """
    stem = path.stem
    if (
        not _IDENTIFIER_RE.match(stem)
        or path.parent == pathlib.Path()
        or path.suffix != ".py"
    ):
        qualified_name = re.sub(r"[^A-Za-z0-9_]", "_", stem)
    else:
        sys_path_set = set(sys_path)
        ancestor_in_sys_path = False

        if path.name == "__init__.py":
            # Go one level up the directory path as the module name comes from the base directory
            reversed_parts = [path.parent.stem]
            ancestors = path.parent.parents
        else:
            # Start from the file as the base directory refers to the parent module
            reversed_parts = [stem]
            ancestors = path.parents

        for ancestor in ancestors:
            if str(ancestor) in sys_path_set:
                ancestor_in_sys_path = True
                break

            # Check to see if the ancestor is a proper Python parent module
            if (
                not _IDENTIFIER_RE.match(ancestor.stem)
                or not (ancestor / "__init__.py").exists()
            ):
                break

            # Continue to see if we can reach one of the paths in sys.path
            reversed_parts.append(ancestor.stem)

        if ancestor_in_sys_path:
            qualified_name = ".".join(reversed(reversed_parts))
        else:
            qualified_name = re.sub(r"[^A-Za-z0-9_]", "_", stem)

    return qualified_name


def ghostwrite(general: _general.Params, command: Params) -> Tuple[str, List[str]]:
    """
    Write a unit test module for the specified functions.

    Return (generated code, errors if any).
    """
    if command.module_name is not None:
        mod, errors = _general.load_module_with_name(command.module_name)
    elif command.path is not None:
        mod, errors = _general.load_module_from_source_file(path=command.path)
    else:
        raise AssertionError(
            f"Unexpected execution path. The command was: {command.__dict__!r}"
        )

    if errors:
        return "", errors

    assert mod is not None

    ##
    # Load the source code
    ##

    if command.module_name is not None:
        source_code = inspect.getsource(mod)
    elif command.path is not None:
        source_code = command.path.read_text()
    else:
        raise AssertionError(
            f"Unexpected execution path. The command was: {command.__dict__!r}"
        )

    ##
    # Select points
    ##

    points, errors = _general.select_function_points(
        source_code=source_code,
        mod=mod,
        include=general.include,
        exclude=general.exclude,
    )
    if errors:
        return "", errors

    ##
    # Figure out the qualified name of the module
    ##

    if command.module_name is not None:
        qualified_name = command.module_name
    elif command.path is not None:
        qualified_name = _qualified_module_name_from_path(
            path=command.path, sys_path=sys.path
        )
    else:
        raise AssertionError(
            f"Unexpected execution path. The command was: {command.__dict__!r}"
        )

    ##
    # Ghostwrite
    ##

    return _ghostwrite_for_function_points(
        points=points,
        module_name=qualified_name,
        explicit=command.explicit,
        bare=command.bare,
    )



"""
Test Python 3.7-specific features.

For example, we could not test ``test_strategy_inference_on_forward_declarations`` in Python 3.6 as
the multiple inheritance of ``icontract.DBC`` and ``Sequance[int]`` caused a meta-class conflict.
"""

import sys

if sys.version_info < (3, 7):
    def load_tests(loader, suite, pattern):  # pylint: disable=unused-argument
        """Ignore all the tests for lower Python versions."""
        return suite



# pylint: disable=missing-docstring
# pylint: disable=invalid-name
# pylint: disable=unused-argument
import io
import os
import pathlib
import re
import shutil
import sys
import tempfile
import unittest
from typing import List

from icontract_hypothesis.pyicontract_hypothesis import _general, _ghostwrite, main


class TestParsingOfParameters(unittest.TestCase):
    def test_subcommand_ghostwrite(self) -> None:
        # fmt: off
        argv = [
            "ghostwrite",
            "--module", "some_module",
            "--include", "include-something",
            "--exclude", "exclude-something",
            "--explicit",
            "--bare",
        ]
        # fmt: on
        parser = main._make_argument_parser()
        args, out, err = main._parse_args(parser=parser, argv=argv)
        assert args is not None, "Failed to parse argv {!r}: {}".format(argv, err)

        general, errs = _general.parse_params(args=args)

        self.assertListEqual([], errs)
        assert general is not None
        self.assertListEqual(
            [re.compile(pattern) for pattern in ["include-something"]], general.include
        )
        self.assertListEqual(
            [re.compile(pattern) for pattern in ["exclude-something"]], general.exclude
        )

        ghostwrite, errs = _ghostwrite.parse_params(args=args)

        self.assertListEqual([], errs)
        assert ghostwrite is not None
        self.assertEqual("some_module", ghostwrite.module_name)
        self.assertTrue(ghostwrite.explicit)
        self.assertTrue(ghostwrite.bare)

    def test_path_as_input(self) -> None:
        # fmt: off
        argv = [
            "ghostwrite",
            "--path", "path/to/some_module.py"
        ]
        # fmt: on
        parser = main._make_argument_parser()
        args, out, err = main._parse_args(parser=parser, argv=argv)
        assert args is not None, "Failed to parse argv {!r}: {}".format(argv, err)

        general, errs = _general.parse_params(args=args)

        self.assertListEqual([], errs)
        assert general is not None

        ghostwrite, errs = _ghostwrite.parse_params(args=args)

        self.assertListEqual([], errs)
        assert ghostwrite is not None
        assert ghostwrite.path is not None
        assert ghostwrite.module_name is None

        self.assertEqual("path/to/some_module.py", ghostwrite.path.as_posix())


class TestQualifiedModuleNameFromPath(unittest.TestCase):
    def test_non_identifier_stem(self) -> None:
        path = pathlib.Path("some_directory") / "some-script.py"

        qualified_name = _ghostwrite._qualified_module_name_from_path(
            path=path, sys_path=[]
        )

        self.assertEqual("some_script", qualified_name)

    def test_no_parent_in_sys_path(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir_str:
            tmpdir = pathlib.Path(tmpdir_str)

            (tmpdir / "some_module").mkdir()
            (tmpdir / "some_module" / "__init__.py").write_text('"""something"""')

            path = tmpdir / "some_module" / "some_submodule.py"
            path.write_text('"""something"""')

            # If tmpdir were added to sys_path, we would have a meaningful qualified name.
            # As sys_path is empty, we have to resort to a stem.
            qualified_name = _ghostwrite._qualified_module_name_from_path(
                path=path, sys_path=[]
            )

            self.assertEqual("some_submodule", qualified_name)

    def test_path_to_module_file(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir_str:
            tmpdir = pathlib.Path(tmpdir_str)

            (tmpdir / "some_module").mkdir()
            (tmpdir / "some_module" / "__init__.py").write_text('"""something"""')

            path = tmpdir / "some_module" / "some_submodule.py"
            path.write_text('"""something"""')

            qualified_name = _ghostwrite._qualified_module_name_from_path(
                path=path, sys_path=[tmpdir_str]
            )

            self.assertEqual("some_module.some_submodule", qualified_name)

    def test_path_to_init(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir_str:
            tmpdir = pathlib.Path(tmpdir_str)

            (tmpdir / "some_module").mkdir()
            (tmpdir / "some_module" / "__init__.py").write_text('"""something"""')

            (tmpdir / "some_module" / "some_submodule").mkdir()

            path = tmpdir / "some_module" / "some_submodule" / "__init__.py"
            path.write_text('"""something"""')

            qualified_name = _ghostwrite._qualified_module_name_from_path(
                path=path, sys_path=[tmpdir_str]
            )

            self.assertEqual("some_module.some_submodule", qualified_name)

    def test_path_to_module_file_with_ancestor_without_init(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir_str:
            tmpdir = pathlib.Path(tmpdir_str)

            (tmpdir / "some_module").mkdir()
            # No __init__.py in the directory ``some_module``.

            path = tmpdir / "some_module" / "some_submodule.py"
            path.write_text('"""something"""')

            qualified_name = _ghostwrite._qualified_module_name_from_path(
                path=path, sys_path=[tmpdir_str]
            )

            self.assertEqual("some_submodule", qualified_name)


class TestGhostwrite(unittest.TestCase):
    def test_path(self) -> None:
        this_dir = pathlib.Path(os.path.realpath(__file__)).parent

        path = (
            this_dir.parent.parent
            / "test_samples/pyicontract_hypothesis/sample_module.py"
        )
        # fmt: off
        argv = [
            "ghostwrite",
            "--path", str(path)
        ]
        # fmt: on

        stdout = io.StringIO()
        stderr = io.StringIO()

        exit_code = main.run(argv=argv, stdout=stdout, stderr=stderr)

        self.assertEqual("", stderr.getvalue())
        self.assertEqual(exit_code, 0)

        # The qualified name of the module depends on the sys.path which, in turn, is modified
        # by the test executor (*e.g.*, IDE can add temporarily directories to sys.path
        # automatically when executing tests).
        #
        # Since the qualified name of a module depends on ``sys.path``, we need to replace it with
        # a generic placeholder to make the test output uniform and reproducible.

        qualified_name = _ghostwrite._qualified_module_name_from_path(
            path=path, sys_path=sys.path
        )
        got = stdout.getvalue()
        got = got.replace(qualified_name, "placeholder_for_sample_module")

        expected_pth = (
            this_dir.parent.parent
            / "test_samples/pyicontract_hypothesis/expected_ghostwrites"
            / ("for_{}.txt".format(TestGhostwrite.test_path.__name__))
        )

        expected = expected_pth.read_text()
        self.assertEqual(expected, got)

    def test_bare_and_explicit(self) -> None:
        # fmt: off
        argv = [
            "ghostwrite",
            "--module", "test_samples.pyicontract_hypothesis.sample_module",
            "--bare",
            "--explicit"
        ]
        # fmt: on

        stdout = io.StringIO()
        stderr = io.StringIO()

        exit_code = main.run(argv=argv, stdout=stdout, stderr=stderr)

        self.assertEqual("", stderr.getvalue())
        self.assertEqual(exit_code, 0)

        this_dir = pathlib.Path(os.path.realpath(__file__)).parent
        expected_pth = (
            this_dir.parent.parent
            / "test_samples/pyicontract_hypothesis/expected_ghostwrites"
            / ("for_{}.txt".format(TestGhostwrite.test_bare_and_explicit.__name__))
        )

        expected = expected_pth.read_text()
        self.assertEqual(expected, stdout.getvalue())

    def test_bare_and_non_explicit(self) -> None:
        # fmt: off
        argv = [
            "ghostwrite",
            "--module", "test_samples.pyicontract_hypothesis.sample_module",
            "--bare",
        ]
        # fmt: on

        stdout = io.StringIO()
        stderr = io.StringIO()

        exit_code = main.run(argv=argv, stdout=stdout, stderr=stderr)

        self.assertEqual("", stderr.getvalue())
        self.assertEqual(exit_code, 0)

        this_dir = pathlib.Path(os.path.realpath(__file__)).parent
        expected_pth = (
            this_dir.parent.parent
            / "test_samples/pyicontract_hypothesis/expected_ghostwrites"
            / ("for_{}.txt".format(TestGhostwrite.test_bare_and_non_explicit.__name__))
        )

        expected = expected_pth.read_text()
        self.assertEqual(expected, stdout.getvalue())

    def test_non_bare_and_explicit(self) -> None:
        # fmt: off
        argv = [
            "ghostwrite",
            "--module", "test_samples.pyicontract_hypothesis.sample_module",
            "--explicit"
        ]
        # fmt: on

        stdout = io.StringIO()
        stderr = io.StringIO()

        exit_code = main.run(argv=argv, stdout=stdout, stderr=stderr)

        self.assertEqual("", stderr.getvalue())
        self.assertEqual(exit_code, 0)

        this_dir = pathlib.Path(os.path.realpath(__file__)).parent
        expected_pth = (
            this_dir.parent.parent
            / "test_samples/pyicontract_hypothesis/expected_ghostwrites"
            / ("for_{}.py".format(TestGhostwrite.test_non_bare_and_explicit.__name__))
        )

        expected = expected_pth.read_text()
        self.assertEqual(expected, stdout.getvalue())

    def test_non_bare_and_non_explicit(self) -> None:
        # fmt: off
        argv = [
            "ghostwrite",
            "--module", "test_samples.pyicontract_hypothesis.sample_module",
        ]
        # fmt: on

        stdout = io.StringIO()
        stderr = io.StringIO()

        exit_code = main.run(argv=argv, stdout=stdout, stderr=stderr)

        self.assertEqual("", stderr.getvalue())
        self.assertEqual(exit_code, 0)

        this_dir = pathlib.Path(os.path.realpath(__file__)).parent
        expected_pth = (
            this_dir.parent.parent
            / "test_samples/pyicontract_hypothesis/expected_ghostwrites"
            / (
                "for_{}.py".format(
                    TestGhostwrite.test_non_bare_and_non_explicit.__name__
                )
            )
        )

        expected = expected_pth.read_text()
        self.assertEqual(expected, stdout.getvalue())

    def test_well_formatted_with_two_arguments(self) -> None:
        # This test is related to the issue:
        # https://github.com/mristin/icontract-hypothesis/issues/29
        # fmt: off
        argv = [
            "ghostwrite",
            "--module",
            "test_samples.pyicontract_hypothesis.well_formatted_with_two_arguments",
            "--explicit"
        ]
        # fmt: on

        stdout = io.StringIO()
        stderr = io.StringIO()

        exit_code = main.run(argv=argv, stdout=stdout, stderr=stderr)

        self.assertEqual("", stderr.getvalue())
        self.assertEqual(exit_code, 0)

        this_dir = pathlib.Path(os.path.realpath(__file__)).parent
        expected_pth = (
            this_dir.parent.parent
            / "test_samples/pyicontract_hypothesis/expected_ghostwrites"
            / (
                "for_{}.py".format(
                    TestGhostwrite.test_well_formatted_with_two_arguments.__name__
                )
            )
        )

        expected = expected_pth.read_text()
        self.assertEqual(expected, stdout.getvalue())


if __name__ == "__main__":
    unittest.main()



"""Provide functions common to all the commands."""
import argparse
import importlib
import importlib.machinery
import inspect
import io
import json
import pathlib
import re
import tokenize
import types
from typing import List, Union, Tuple, Optional, Callable, Any, Pattern

import icontract


class _LineRange:
    """Represent a line range (indexed from 1, both first and last inclusive)."""

    def __init__(self, first: int, last: int) -> None:
        """Initialize with the given values."""
        self.first = first
        self.last = last


class Params:
    """Represent general program parameters specified regardless of the command."""

    def __init__(
        self,
        include: List[Union[Pattern[str], _LineRange]],
        exclude: List[Union[Pattern[str], _LineRange]],
    ) -> None:
        """Initialize with the given values."""
        self.include = include  # type: List[Union[Pattern[str], _LineRange]]
        self.exclude = exclude  # type: List[Union[Pattern[str], _LineRange]]


_LINE_RANGE_RE = re.compile(
    r"^\s*(?P<first>[0-9]|[1-9][0-9]+)(\s*-\s*(?P<last>[1-9]|[1-9][0-9]+))?\s*$"
)


def _parse_point_spec(
    text: str,
) -> Tuple[Optional[Union[_LineRange, Pattern[str]]], List[str]]:
    """
    Try to parse the given specification of function point(s).

    Return (parsed point spec, errors if any)
    """
    errors = []  # type: List[str]

    mtch = _LINE_RANGE_RE.match(text)
    if mtch:
        if mtch.group("last") is None:
            first = int(mtch.group("first"))
            if first <= 0:
                errors.append(
                    "Unexpected line index (expected to start from 1): {}".format(text)
                )
                return None, errors

            return _LineRange(first=int(mtch.group("first")), last=first), errors
        else:
            first = int(mtch.group("first"))
            last = int(mtch.group("last"))

            if first <= 0:
                errors.append(
                    "Unexpected line index (expected to start from 1): {}".format(text)
                )
                return None, errors

            if last < first:
                errors.append("Unexpected line range (last < first): {}".format(text))
                return None, errors

            else:
                return (
                    _LineRange(
                        first=int(mtch.group("first")), last=int(mtch.group("last"))
                    ),
                    errors,
                )

    try:
        pattern = re.compile(text)
        return pattern, errors
    except re.error as err:
        errors.append("Failed to parse the pattern {}: {}".format(text, err))
        return None, errors


def parse_params(args: argparse.Namespace) -> Tuple[Optional[Params], List[str]]:
    """
    Try to parse general parameters of the program (regardless of the command).

    Return (parsed parameters, errors if any).
    """
    errors = []  # type: List[str]

    include = []  # type: List[Union[Pattern[str], _LineRange]]
    if args.include is not None:
        for include_str in args.include:
            point_spec, point_spec_errors = _parse_point_spec(text=include_str)
            errors.extend(point_spec_errors)

            if not point_spec_errors:
                assert point_spec is not None
                include.append(point_spec)

    exclude = []  # type: List[Union[Pattern[str], _LineRange]]
    if args.exclude is not None:
        for exclude_str in args.exclude:
            point_spec, point_spec_errors = _parse_point_spec(text=exclude_str)
            errors.extend(point_spec_errors)

            if not point_spec_errors:
                assert point_spec is not None
                exclude.append(point_spec)

    if errors:
        return None, errors

    return Params(include=include, exclude=exclude), errors


def load_module_with_name(name: str) -> Tuple[Optional[types.ModuleType], List[str]]:
    """
    Load the module given its name.

    Example identifier: some.module
    """
    try:
        mod = importlib.import_module(name=name)
        assert isinstance(mod, types.ModuleType)
        return mod, []
    except Exception as error:
        return None, ["Failed to import the module {}: {}".format(name, error)]


def load_module_from_source_file(
    path: pathlib.Path,
) -> Tuple[Optional[types.ModuleType], List[str]]:
    """
    Try to load a module from the source file.

    Return (loaded module, errors if any).
    """
    fullname = re.sub(r"[^A-Za-z0-9_]", "_", path.stem)

    mod = None  # type: Optional[types.ModuleType]
    try:
        loader = importlib.machinery.SourceFileLoader(fullname=fullname, path=str(path))
        mod = types.ModuleType(loader.name)
        loader.exec_module(mod)
    except Exception as error:
        return None, ["Failed to import the file {}: {}".format(path, error)]

    assert mod is not None, "Expected mod to be set before"

    return mod, []


class FunctionPoint:
    """Represent a testable function."""

    @icontract.require(lambda first_row: first_row > 0)
    @icontract.require(lambda last_row: last_row > 0)
    @icontract.require(lambda first_row, last_row: first_row <= last_row)
    def __init__(self, first_row: int, last_row: int, func: Callable[..., Any]) -> None:
        """
        Initialize with the given values.

        First and last row are both inclusive.
        """
        self.first_row = first_row
        self.last_row = last_row
        self.func = func


def _overlap(first: int, last: int, another_first: int, another_last: int) -> bool:
    """
    Return True if the two intervals overlap.

    >>> not any([
    ...     _overlap(1, 1, 2, 2),
    ...     _overlap(2, 2, 1, 1)
    ... ])
    True

    >>> all([
    ...     _overlap(1, 1, 1, 1),
    ...     _overlap(1, 5, 1, 1),
    ...     _overlap(1, 1, 1, 5),
    ...     _overlap(1, 3, 2, 5),
    ...     _overlap(2, 5, 1, 3),
    ...     _overlap(1, 5, 2, 3),
    ...     _overlap(2, 3, 1, 5),
    ...  ])
    True
    """
    return min(last, another_last) - max(first, another_first) >= 0


_DIRECTIVE_RE = re.compile(r"^#\s*pyicontract-hypothesis\s*:\s*(?P<value>[^ \t]*)\s*$")


def select_function_points(
    source_code: str,
    mod: types.ModuleType,
    include: List[Union[_LineRange, Pattern[str]]],
    exclude: List[Union[_LineRange, Pattern[str]]],
) -> Tuple[List[FunctionPoint], List[str]]:
    """Select the function points from the module based on the ``include`` and ``exclude``."""
    included = []  # type: List[FunctionPoint]
    errors = []  # type: List[str]

    for key in dir(mod):
        value = getattr(mod, key)
        if inspect.isfunction(value):
            func = value  # type: Callable[..., Any]

            # Ignore imported functions
            if func.__module__ != mod.__name__:
                continue

            source_lines, srow = inspect.getsourcelines(func)
            point = FunctionPoint(
                first_row=srow, last_row=srow + len(source_lines) - 1, func=func
            )
            included.append(point)

    # The built-in dir() gives us an unsorted directory.
    included = sorted(included, key=lambda point: point.first_row)

    ##
    # Add ranges of lines given by comment directives to the ``exclude``
    ##

    extended_exclude = exclude[:]

    range_start = None  # type: Optional[int]
    reader = io.BytesIO(source_code.encode("utf-8"))
    for toktype, _, (first_row, _), _, line in tokenize.tokenize(reader.readline):
        if toktype == tokenize.COMMENT:
            mtch = _DIRECTIVE_RE.match(line.strip())
            if mtch:
                value = mtch.group("value")

                if value not in ["enable", "disable"]:
                    errors.append(
                        (
                            "Unexpected directive on line {}. "
                            "Expected '# pyicontract-hypothesis: (disable|enable)', "
                            "but got: {}"
                        ).format(first_row, line.strip())
                    )
                    continue

                if value == "disable":
                    if range_start is not None:
                        continue

                    range_start = first_row

                elif value == "enable":
                    if range_start is not None:
                        extended_exclude.append(
                            _LineRange(first=range_start, last=first_row)
                        )

                else:
                    raise AssertionError(
                        "Unexpected value: {}".format(json.dumps(value))
                    )

    exclude = extended_exclude

    if errors:
        return [], errors

    ##
    # Remove ``included`` which do not match ``include``
    ##

    if len(include) > 0:
        incl_line_ranges = [incl for incl in include if isinstance(incl, _LineRange)]
        if len(incl_line_ranges) > 100:
            print(
                (
                    "There are much more --include items then expected: {0}. "
                    "Please consider filing an issue by visiting this link: "
                    "https://github.com/Parquery/icontract/issues/new"
                    "?title=Use+interval+tree"
                    "&body=We+had+{0}+include+line+ranges+in+pyicontract-hypothesis."
                ).format(len(incl_line_ranges))
            )

        if len(incl_line_ranges) > 0:
            filtered_included = []  # type: List[FunctionPoint]
            for point in included:
                overlaps_include = any(
                    _overlap(
                        first=line_range.first,
                        last=line_range.last,
                        another_first=point.first_row,
                        another_last=point.last_row,
                    )
                    for line_range in incl_line_ranges
                )

                if overlaps_include:
                    filtered_included.append(point)

            included = filtered_included

        # Match regular expressions
        patterns = [incl for incl in include if isinstance(incl, Pattern)]
        if len(patterns) > 0:
            filtered_included = []
            for pattern in patterns:
                for point in included:
                    if pattern.match(point.func.__name__):
                        filtered_included.append(point)

            included = filtered_included

    if len(included) == 0:
        return [], []

    ##
    # Exclude all points in ``included`` if matched in ``exclude``
    ##

    if len(exclude) > 0:
        excl_line_ranges = [excl for excl in exclude if isinstance(excl, _LineRange)]
        if len(excl_line_ranges) > 100:
            print(
                (
                    "There are much more --exclude items then expected: {0}. "
                    "Please consider filing an issue by visiting this link: "
                    "https://github.com/Parquery/icontract/issues/new"
                    "?title=Use+interval+tree"
                    "&body=We+had+{0}+exclude+line+ranges+in+pyicontract-hypothesis."
                ).format(len(excl_line_ranges))
            )

        if len(excl_line_ranges) > 0:
            filtered_included = []
            for point in included:
                overlaps_exclude = any(
                    _overlap(
                        first=line_range.first,
                        last=line_range.last,
                        another_first=point.first_row,
                        another_last=point.last_row,
                    )
                    for line_range in excl_line_ranges
                )

                if not overlaps_exclude:
                    filtered_included.append(point)

            included = filtered_included

        patterns = [excl for excl in exclude if isinstance(excl, Pattern)]
        if len(patterns) > 0:
            filtered_included = []
            for pattern in patterns:
                for point in included:
                    if not pattern.match(point.func.__name__):
                        filtered_included.append(point)

            included = filtered_included

    return included, []



"""
Integrate icontract with Hypothesis.

You need to install ``hypothesis`` extras in order to use this module.
"""
# pylint: disable=unused-argument
# pylint: disable=protected-access
# pylint: disable=inconsistent-return-statements
# pylint: disable=too-many-lines

import ast
import datetime
import decimal
import dis
import fractions
import inspect
import io
import re
import sys
import types
import typing
from typing import (
    cast,
    TypeVar,
    Callable,
    Any,
    Generic,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Union,
    Dict,
    Tuple,
    Type,
    AnyStr,
    Pattern,
    Set,
)

import hypothesis.errors
import hypothesis.internal.reflection
import hypothesis.strategies
import hypothesis.strategies._internal.collections
import hypothesis.strategies._internal.types
import icontract._checkers
import icontract._metaclass
import icontract._recompute
import icontract._represent
import icontract._types

# Don't forget to update the version in setup.py!
__version__ = "1.1.4"
__author__ = "Marko Ristin"
__copyright__ = "Copyright 2020-2021 Marko Ristin"
__license__ = "MIT"
__status__ = "Production/Stable"

CallableT = TypeVar("CallableT", bound=Callable[..., Any])

T = TypeVar("T")  # pylint: disable=invalid-name


def _assume_preconditions_always_satisfied(
    *args: Tuple[Any, ...], **kwargs: Dict[str, Any]
) -> None:
    """Assume that the preconditions are always satisfied (say, when there are no preconditions)."""


def make_assume_preconditions(func: CallableT) -> Callable[..., None]:
    """
    Create a function that assumes the preconditions are satisfied given the positional and keyword arguments.

    Here is an example test case which tests a function ``some_func``:

    >>> import unittest
    >>> import hypothesis.strategies as st
    >>> import icontract_hypothesis
    >>> import icontract

    >>> @icontract.require(lambda x: x > 0)
    ... def some_func(x: int) -> None:
    ...     ...

    >>> class TestSomething(unittest.TestCase):
    ...     def test_that_it_works(self) -> None:
    ...         assume_preconditions = icontract_hypothesis.make_assume_preconditions(
    ...             some_func)
    ...
    ...         @hypothesis.given(x=st.integers())
    ...         def run(x: int) -> None:
    ...             assume_preconditions(x)
    ...             some_func(x)
    ...         run()

    >>> TestSomething().test_that_it_works()
    """
    # The implementation follows tightly icontract._checkers.decorate_with_checker and
    # icontract._checkers._assert_precondition.

    checker = icontract._checkers.find_checker(func)
    if checker is None:
        return _assume_preconditions_always_satisfied

    preconditions = getattr(checker, "__preconditions__", None)
    if preconditions is None:
        return _assume_preconditions_always_satisfied

    sign = inspect.signature(func)
    param_names = list(sign.parameters.keys())
    kwdefaults = icontract._checkers.resolve_kwdefaults(sign=sign)

    def assume_preconditions(*args, **kwargs) -> None:  # type: ignore
        """Accept only positional and keyword arguments that satisfy one of the precondition groups."""
        resolved_kwargs = icontract._checkers.kwargs_from_call(
            param_names=param_names, kwdefaults=kwdefaults, args=args, kwargs=kwargs
        )

        success = True

        for group in preconditions:  # pylint: disable=not-an-iterable
            success = True
            for contract in group:
                condition_kwargs = icontract._checkers.select_condition_kwargs(
                    contract=contract, resolved_kwargs=resolved_kwargs
                )

                check = contract.condition(**condition_kwargs)

                if icontract._checkers.not_check(check=check, contract=contract):
                    success = False
                    break

            if success:
                return

        if not success:
            raise hypothesis.errors.UnsatisfiedAssumption()

    return assume_preconditions


class _InferredMinMax:
    """Represent the inference result of boundaries on an argument."""

    def __init__(
        self,
        min_value: Optional[Any] = None,
        min_inclusive: bool = False,
        max_value: Optional[Any] = None,
        max_inclusive: bool = False,
    ) -> None:
        """Initialize with the given values."""
        self.min_value = min_value
        self.min_inclusive = min_inclusive
        self.max_value = max_value
        self.max_inclusive = max_inclusive


def _no_name_in_descendants(root: ast.expr, name: str) -> bool:
    """Check whether a ``ast.Name`` node with ``root`` identifier is present in the descendants of the node."""
    found = False

    class Visitor(ast.NodeVisitor):
        """Search for the name node."""

        def visit_Name(  # pylint: disable=invalid-name,no-self-use,missing-docstring
            self, node: ast.Name
        ) -> None:
            if node.id == name:
                nonlocal found
                found = True

        def generic_visit(self, node: Any) -> None:
            if not found:
                super(Visitor, self).generic_visit(node)

    visitor = Visitor()
    visitor.visit(root)

    return not found


def _recompute(condition: Callable[..., Any], node: ast.expr) -> Tuple[Any, bool]:
    """Recompute the value corresponding to the node."""
    recompute_visitor = icontract._recompute.Visitor(
        variable_lookup=icontract._represent.collect_variable_lookup(
            condition=condition, resolved_kwargs=None
        )
    )

    recompute_visitor.visit(node=node)

    if node in recompute_visitor.recomputed_values:
        return recompute_visitor.recomputed_values[node], True

    return None, False


def _infer_min_max_from_node(
    condition: Callable[..., bool], node: ast.Compare, arg_name: str
) -> Optional[_InferredMinMax]:
    """Match one of the patterns against the AST compare node."""
    # pylint: disable=too-many-boolean-expressions
    # pylint: disable=too-many-return-statements
    # pylint: disable=too-many-branches
    if len(node.comparators) == 1:
        comparator = node.comparators[0]
        operation = node.ops[0]

        # Match something like "x > 0" and "x < 100"
        if (
            isinstance(node.left, ast.Name)
            and node.left.id == arg_name
            and _no_name_in_descendants(root=comparator, name=arg_name)
        ):
            value, recomputed = _recompute(condition=condition, node=comparator)

            # If we can not recompute the value, we also can not infer the bounds.
            if not recomputed:
                return None

            # Match something like "x < 100"
            if isinstance(operation, ast.Lt):
                return _InferredMinMax(max_value=value, max_inclusive=False)
            # Match something like "x =< 100"
            elif isinstance(operation, ast.LtE):
                return _InferredMinMax(max_value=value, max_inclusive=True)

            # Match something like "x > 0"
            elif isinstance(operation, ast.Gt):
                return _InferredMinMax(min_value=value, min_inclusive=False)
            # Match something like "x >= 0"
            elif isinstance(operation, ast.GtE):
                return _InferredMinMax(min_value=value, min_inclusive=True)

            # We could not infer any bound from this condition.
            else:
                return None

        # Match something like "0 < x" and "100 > x"
        if (
            _no_name_in_descendants(root=node.left, name=arg_name)
            and isinstance(comparator, ast.Name)
            and comparator.id == arg_name
        ):
            value, recomputed = _recompute(condition=condition, node=node.left)

            # If we can not recompute the value, we also can not infer the bounds.
            if not recomputed:
                return None

            # Match something like "0 < x"
            if isinstance(operation, ast.Lt):
                return _InferredMinMax(min_value=value, min_inclusive=False)

            # Match something like "0 =< x"
            if isinstance(operation, ast.LtE):
                return _InferredMinMax(min_value=value, min_inclusive=True)

            # Match something like "100 > x"
            elif isinstance(operation, ast.Gt):
                return _InferredMinMax(max_value=value, max_inclusive=False)

            # Match something like "100 >= x"
            elif isinstance(operation, ast.GtE):
                return _InferredMinMax(max_value=value, max_inclusive=True)

            # We could not infer any bound from this condition.
            else:
                pass

    elif len(node.comparators) == 2:
        # Match something like "0 < x < 100" and "0 > x > -100"
        if (
            _no_name_in_descendants(root=node.left, name=arg_name)
            and isinstance(node.comparators[0], ast.Name)
            and node.comparators[0].id == arg_name
            and _no_name_in_descendants(root=node.comparators[1], name=arg_name)
        ):

            left_value, recomputed = _recompute(condition=condition, node=node.left)

            # If we can not recompute the left value, we also can not infer the bounds.
            if not recomputed:
                return None

            right_value, recomputed = _recompute(
                condition=condition, node=node.comparators[1]
            )

            # If we can not recompute the right value, we also can not infer the bounds.
            if not recomputed:
                return None

            op0, op1 = node.ops

            # Match something like "0 < x < 100"
            if isinstance(op0, (ast.Lt, ast.LtE)) and isinstance(
                op1, (ast.Lt, ast.LtE)
            ):
                return _InferredMinMax(
                    min_value=left_value,
                    min_inclusive=isinstance(op0, ast.LtE),
                    max_value=right_value,
                    max_inclusive=isinstance(op1, ast.LtE),
                )

            # Match something like "0 > x > -100"
            elif isinstance(op0, (ast.Gt, ast.GtE)) and isinstance(
                op1, (ast.Gt, ast.GtE)
            ):
                return _InferredMinMax(
                    min_value=right_value,
                    min_inclusive=isinstance(op0, ast.GtE),
                    max_value=left_value,
                    max_inclusive=isinstance(op1, ast.GtE),
                )

            # We could not infer any bound from this condition.
            else:
                pass

    return None


def _body_node_from_condition(condition: Callable[..., Any]) -> Optional[ast.expr]:
    """Try to extract the body node of the contract's lambda condition."""
    if not icontract._represent.is_lambda(a_function=condition):
        return None

    lines, condition_lineno = inspect.findsource(condition)
    filename = inspect.getsourcefile(condition)
    assert filename is not None

    decorator_inspection = icontract._represent.inspect_decorator(
        lines=lines, lineno=condition_lineno, filename=filename
    )
    lambda_inspection = icontract._represent.find_lambda_condition(
        decorator_inspection=decorator_inspection
    )

    assert (
        lambda_inspection is not None
    ), "Expected lambda_inspection to be non-None if _is_lambda is True on: {}".format(
        condition
    )

    body_node = lambda_inspection.node.body

    return body_node


@icontract.require(
    lambda arg_name, contracts: all(
        len(contract.condition_args) == 1 and contract.condition_args[0] == arg_name
        for contract in contracts
    ),
    "All contracts are single-argument contracts related to this argument.",
)
def _infer_min_max_from_preconditions(
    arg_name: str, type_hint: Type[T], contracts: List[icontract._types.Contract]
) -> Tuple[_InferredMinMax, List[icontract._types.Contract]]:
    """
    Infer the min and max values for the given argument from all related preconditions.

    Return the contracts which could not be interpreted.
    """
    min_value = None  # type: Optional[Union[int, float]]
    max_value = None  # type: Optional[Union[int, float]]

    remaining_contracts = []  # type: List[icontract._types.Contract]

    for contract in contracts:
        body_node = _body_node_from_condition(condition=contract.condition)

        if body_node is None:
            remaining_contracts.append(contract)
            continue

        if isinstance(body_node, ast.Compare):
            inferred = _infer_min_max_from_node(
                condition=contract.condition, node=body_node, arg_name=arg_name
            )

            if inferred is not None:
                # We need to constrain min and max values.
                # Hence we use ``max`` for min and ``min`` for max, respectively.
                # This might be a bit counter-intuitive at the first sight.

                if inferred.min_value is not None:
                    min_value = (
                        inferred.min_value
                        if min_value is None
                        else max(inferred.min_value, min_value)
                    )

                if inferred.max_value is not None:
                    max_value = (
                        inferred.max_value
                        if max_value is None
                        else min(inferred.max_value, max_value)
                    )

                # We can not have exclusive bounds on Fractions and Decimals, so we need to leave
                # the contracts in place.
                if type_hint in [fractions.Fraction, decimal.Decimal]:
                    remaining_contracts.append(contract)
            else:
                remaining_contracts.append(contract)
        else:
            remaining_contracts.append(contract)

    return (
        _InferredMinMax(min_value=min_value, max_value=max_value),
        remaining_contracts,
    )


def _make_strategy_with_min_max_for_type(
    a_type: Type[T], inferred: _InferredMinMax
) -> hypothesis.strategies.SearchStrategy[T]:
    if a_type == int:
        # hypothesis.strategies.integers is always inclusive so we have to cut off the boundaries
        # a bit if they are exclusive.
        min_value = inferred.min_value
        if min_value is not None and not inferred.min_inclusive:
            min_value += 1

        max_value = inferred.max_value
        if max_value is not None and not inferred.max_inclusive:
            max_value -= 1

        strategy = hypothesis.strategies.integers(
            min_value=min_value,
            max_value=max_value,
        )  # type: hypothesis.strategies.SearchStrategy[Any]

    elif a_type == float:
        strategy = hypothesis.strategies.floats(
            min_value=inferred.min_value,
            max_value=inferred.max_value,
            exclude_min=inferred.min_value is not None and not inferred.min_inclusive,
            exclude_max=inferred.max_value is not None and not inferred.max_inclusive,
        )

    elif a_type == fractions.Fraction:
        # Fractions must also include their contracts as we can not bound them
        # by filtering! This needs to be a special case.
        strategy = hypothesis.strategies.fractions(
            min_value=inferred.min_value, max_value=inferred.max_value
        )

    elif a_type == decimal.Decimal:
        # Decimals must also include their contracts as we can not bound them
        # by filtering! This needs to be a special case.
        strategy = hypothesis.strategies.decimals(
            min_value=inferred.min_value, max_value=inferred.max_value
        )

    elif a_type == datetime.date:
        min_value = datetime.date.min
        if inferred.min_value is not None:
            if inferred.min_inclusive:
                min_value = inferred.min_value
            else:
                min_value = inferred.min_value + datetime.timedelta(days=1)

        max_value = datetime.date.max
        if inferred.max_value is not None:
            if inferred.max_inclusive:
                max_value = inferred.max_value
            else:
                max_value = inferred.max_value - datetime.timedelta(days=1)

        strategy = hypothesis.strategies.dates(min_value=min_value, max_value=max_value)

    elif a_type == datetime.datetime:
        min_value = datetime.datetime.min
        if inferred.min_value is not None:
            if inferred.min_inclusive:
                min_value = inferred.min_value
            else:
                min_value = inferred.min_value + datetime.timedelta(microseconds=1)

        max_value = datetime.datetime.max
        if inferred.max_value is not None:
            if inferred.max_inclusive:
                max_value = inferred.max_value
            else:
                max_value = inferred.max_value - datetime.timedelta(microseconds=1)

        strategy = hypothesis.strategies.datetimes(
            min_value=min_value, max_value=max_value
        )

    elif a_type == datetime.time:
        min_value = datetime.time.min
        if inferred.min_value is not None:
            min_value = inferred.min_value

            if not inferred.min_inclusive:
                if min_value == datetime.time.max:
                    raise ValueError(
                        f"The inferred exclusive lower bound for the time is equal "
                        f"datetime.time.max ({datetime.time.max}) "
                        f"so we can not compute the next greater time."
                    )

                min_value = (
                    datetime.datetime.combine(
                        datetime.date(2000, 1, 1), min_value, min_value.tzinfo
                    )
                    + datetime.timedelta(microseconds=1)
                ).time()

        max_value = datetime.time.max
        if inferred.max_value is not None:
            max_value = inferred.max_value

            if not inferred.max_inclusive:
                if max_value == datetime.time.min:
                    raise ValueError(
                        f"The inferred exclusive upper bound for the time is equal "
                        f"datetime.time.min ({datetime.time.min}) "
                        f"so we can not compute the previous less-than time."
                    )

                max_value = (
                    datetime.datetime.combine(
                        datetime.date(2000, 1, 1), max_value, max_value.tzinfo
                    )
                    - datetime.timedelta(microseconds=1)
                ).time()

        strategy = hypothesis.strategies.times(min_value=min_value, max_value=max_value)

    elif a_type == datetime.timedelta:
        min_value = datetime.timedelta.min
        if inferred.min_value is not None:
            if inferred.min_inclusive:
                min_value = inferred.min_value
            else:
                min_value = inferred.min_value + datetime.timedelta(microseconds=1)

        max_value = datetime.timedelta.max
        if inferred.max_value is not None:
            if inferred.max_inclusive:
                max_value = inferred.max_value
            else:
                max_value = inferred.max_value - datetime.timedelta(microseconds=1)

        strategy = hypothesis.strategies.timedeltas(
            min_value=min_value, max_value=max_value
        )

    else:
        raise AssertionError("Unexpected type hint: {}".format(a_type))

    return strategy


# We need to compile a dummy pattern so that we can compare addresses of re.Pattern.match functions.
_DUMMY_RE = re.compile(r"something")


class PatternInCondition(Generic[AnyStr]):
    """Represent a pattern that defines the contract condition."""

    def __init__(self, pattern: Pattern[AnyStr], fullmatch: bool) -> None:
        """Initialize with the given values."""
        self.pattern: Pattern[AnyStr] = pattern
        self.fullmatch = fullmatch


def _infer_regexp_from_condition(
    arg_name: str, condition: Callable[..., Any]
) -> Optional[PatternInCondition[AnyStr]]:
    """Try to infer the regular expression pattern from a precondition."""
    body_node = _body_node_from_condition(condition=condition)
    if body_node is None:
        return None

    if not isinstance(body_node, ast.Call):
        return None

    if not _no_name_in_descendants(root=body_node.func, name=arg_name):
        return None

    if not isinstance(body_node.func, ast.Attribute):
        return None

    if body_node.func.attr == "match":
        fullmatch = False
    elif body_node.func.attr == "fullmatch":
        fullmatch = True
    else:
        # This is neither ``re.match`` nor ``re.fullmatch`` function.
        return None

    callee, recomputed = _recompute(condition=condition, node=body_node.func.value)
    if not recomputed:
        return None

    if callee == re:
        # Match "re.match(r'Some pattern', s, *args, *kwargs)
        if (
            len(body_node.args) >= 2
            and _no_name_in_descendants(root=body_node.args[0], name=arg_name)
            and isinstance(body_node.args[1], ast.Name)
            and body_node.args[1].id == arg_name
            and not any(
                _no_name_in_descendants(root=arg, name=arg_name)
                for arg in body_node.args[2:]
            )
        ):
            # Recompute the pattern
            args = []  # type: List[Any]
            for arg in [body_node.args[0]] + body_node.args[2:]:
                value, recomputed = _recompute(condition=condition, node=arg)
                if not recomputed:
                    return None

                args.append(value)

            kwargs = dict()  # type: Dict[str, Any]
            for keyword in body_node.keywords:
                value, recomputed = _recompute(condition=condition, node=keyword.value)
                if not recomputed:
                    return None

                assert (
                    keyword.arg is not None
                ), "Unexpected missing arg for a keyword: {}".format(ast.dump(keyword))
                kwargs[keyword.arg] = value

            pattern = re.compile(*args, **kwargs)
            return PatternInCondition(pattern=pattern, fullmatch=fullmatch)

    elif isinstance(callee, Pattern):
        return PatternInCondition(pattern=callee, fullmatch=fullmatch)

    return None


@icontract.require(
    lambda arg_name, contracts: all(
        len(contract.condition_args) == 1 and contract.condition_args[0] == arg_name
        for contract in contracts
    ),
    "All contracts are single-argument contracts related to this argument.",
)
def _infer_str_strategy_from_preconditions(
    arg_name: str, contracts: List[icontract._types.Contract]
) -> Tuple[
    Optional[hypothesis.strategies.SearchStrategy[AnyStr]],
    List[icontract._types.Contract],
]:
    """
    Try to match code patterns on AST of the preconditions contracts and infer the string strategy.

    Return (strategy if possible, remaining contracts).
    """
    found_idx = -1  # Index of the contract that defines the pattern, -1 if not found
    pattern_in_condition = None  # type: Optional[PatternInCondition[AnyStr]]

    for i, contract in enumerate(contracts):
        pattern_in_condition = _infer_regexp_from_condition(
            arg_name=arg_name, condition=contract.condition
        )
        if pattern_in_condition is not None:
            found_idx = i
            break

    if found_idx == -1:
        return None, contracts[:]

    assert pattern_in_condition is not None
    return (
        hypothesis.strategies.from_regex(
            regex=pattern_in_condition.pattern, fullmatch=pattern_in_condition.fullmatch
        ),
        contracts[:found_idx] + contracts[found_idx + 1 :],
    )


def _strategy_for_type(
    a_type: Type[T],
) -> hypothesis.strategies.SearchStrategy[T]:
    """Create a strategy for instances to satisfy the preconditions on ``__init__``."""
    init = getattr(a_type, "__init__")

    if inspect.isfunction(init):
        strategy = infer_strategy(init)
    elif isinstance(init, icontract._checkers._SLOT_WRAPPER_TYPE):
        # We have to distinguish this special case which is used by named tuples and
        # possibly other optimized data structures.
        # In those cases, we have to infer the strategy based on __new__ instead of __init__.
        new = getattr(a_type, "__new__")
        assert (
            new is not None
        ), "Expected __new__ in {} if __init__ is a slot wrapper.".format(a_type)

        # Add a local namespace in case there are forward references (which is usually the case for
        # the return value).
        #
        # In particular, the class is not available in the module while we are
        # registering it through the ``icontract.DBCMeta`` meta-class as its loading is still
        # in progress.
        strategy = infer_strategy(new, localns={a_type.__name__: a_type})
    else:
        raise AssertionError(
            "Expected __init__ to be either a function or a slot wrapper, but got: {}".format(
                type(init)
            )
        )

    pack_repr = f"lambda d: {a_type.__name__}(**d)"

    pack = lambda d: a_type(**d)  # type: ignore
    pack.__icontract_hypothesis_source_code__ = pack_repr  # type: ignore

    return strategy.map(pack=pack)


@icontract.require(
    lambda arg_name, contracts: contracts is None
    or all(
        len(contract.condition_args) == 1 and contract.condition_args[0] == arg_name
        for contract in contracts
    )
)
def _infer_strategy_for_argument(
    arg_name: str, type_hint: Any, contracts: Optional[List[icontract._types.Contract]]
) -> hypothesis.strategies.SearchStrategy[Any]:
    """Infer the initial strategy for the argument."""
    if contracts is None:
        return hypothesis.strategies.from_type(type_hint)

    strategy = None  # type: Optional[hypothesis.strategies.SearchStrategy[Any]]
    remaining_contracts = contracts

    if type_hint in [
        int,
        float,
        fractions.Fraction,
        decimal.Decimal,
        datetime.date,
        datetime.datetime,
        datetime.time,
        datetime.timedelta,
    ]:
        inferred, remaining_contracts = _infer_min_max_from_preconditions(
            arg_name=arg_name, type_hint=type_hint, contracts=contracts
        )

        if (
            inferred.min_value is not None
            and inferred.max_value is not None
            and inferred.min_value > inferred.max_value
        ):
            raise ValueError(
                (
                    "The min and max values inferred for the argument {} could not be satisfied: "
                    "inferred min is {}, inferred max is {}. Are your preconditions correct?"
                ).format(arg_name, inferred.min_value, inferred.max_value)
            )

        strategy = _make_strategy_with_min_max_for_type(
            a_type=type_hint, inferred=inferred
        )

    if strategy is None and type_hint == str:
        strategy, remaining_contracts = _infer_str_strategy_from_preconditions(
            arg_name=arg_name, contracts=contracts
        )

    if strategy is None:
        strategy = hypothesis.strategies.from_type(type_hint)
        remaining_contracts = contracts

    assert strategy is not None

    for contract in remaining_contracts:
        # Since ``hypothesis.internal.reflection.extract_lambda_source`` is quite
        # limited (as of 2021-04-01), we need to replace the representation of
        # the condition with our own source code.
        #
        # For example, a multi-line condition such as ``lambda lst: all(...)`` will be
        # rendered as ``lambda lst: <unknown>``.
        lambda_inspection = icontract._represent.inspect_lambda_condition(
            condition=contract.condition
        )

        if lambda_inspection is not None:
            # As this attribute is non-invasive, we polute the original condition.
            if not hasattr(contract.condition, "__icontract_hypothesis_source_code__"):
                condition_repr = lambda_inspection.atok.get_text(lambda_inspection.node)
                setattr(
                    contract.condition,
                    "__icontract_hypothesis_source_code__",
                    condition_repr,
                )

        strategy = strategy.filter(contract.condition)

    return strategy


# fmt: off
@icontract.ensure(
    lambda result:
    hasattr(result, "__icontract_hypothesis_source_code__"),
    "The attribute is available (we depend on monkey-patching Hypothesis reflection function)"
)
@icontract.ensure(
    lambda result:
    result.__name__ == "<lambda>",
    "Condition-as-filter-function is lambda "
    "(even in cases where it calls a function as it is rewritten)"
)
# fmt: on
def _rewrite_condition_as_filter_on_kwargs(
    contract: icontract._types.Contract,
) -> Callable[..., Any]:
    """
    Parse, rewrite and recompile the condition so that it can be used as filter on kwargs.

    Return (rewritten condition as filter function, string representation of the condition)
    """
    if "_ARGS" in contract.condition_arg_set:
        raise NotImplementedError(
            "The handling of the special argument _ARGS is not currently handled "
            "by icontract-hypothesis. Please create a new issue on GitHub if you need "
            "this feature: https://github.com/mristin/icontract-hypothesis/issues"
        )

    lambda_inspection = icontract._represent.inspect_lambda_condition(
        condition=contract.condition
    )

    if lambda_inspection is None:
        condition_repr = "lambda d: {}({})".format(
            contract.condition.__name__,
            ", ".join(
                "_KWARGS=d"
                if arg_name == "_KWARGS"
                else "{0}=d['{0}']".format(arg_name)
                for arg_name in contract.condition_args
            ),
        )

        recompiled_code = compile(
            condition_repr,
            filename=f"<rewritten by icontract-hypothesis: {condition_repr}>",
            mode="eval",
        )  # type: types.CodeType

        lambda_code = recompiled_code.co_consts[0]

        rewired_function = types.FunctionType(
            lambda_code,
            {contract.condition.__name__: contract.condition},
            name="<lambda>",
            argdefs=contract.condition.__defaults__,  # type: ignore
            closure=contract.condition.__closure__,  # type: ignore
        )

        rewired_function.__icontract_hypothesis_source_code__ = condition_repr  # type: ignore
        return rewired_function

    name_set = set()  # type: Set[str]

    class FindNamesVisitor(ast.NodeVisitor):
        """Find all the Name nodes."""

        def visit_Name(  # pylint: disable=invalid-name,no-self-use,missing-docstring
            self, node: ast.Name
        ) -> None:
            name_set.add(node.id)

    FindNamesVisitor().visit(lambda_inspection.node)

    # We need to make sure that ``kwargs_name`` does not conflict with any global or closured
    # variable. However, it is OK if it conflicts with the local variables as they are going
    # to be replaced.
    kwargs_name = "d"
    while kwargs_name in name_set.difference(contract.condition_arg_set):
        kwargs_name = "_" + kwargs_name

    class Replace:
        """Instruct to replace the start and end of the source code with the given replacement."""

        def __init__(self, start: int, end: int, replacement: str) -> None:
            self.start = start
            self.end = end
            self.replacement = replacement

        def __repr__(self) -> str:
            return "Replace(start={}, end={}, replacement={!r})".format(
                self.start, self.end, self.replacement
            )

    replacements_in_text = []  # type: List[Replace]

    class ReplaceNameWithSubscriptTransformer(ast.NodeTransformer):
        """Replace condition arguments with kwargs subscript."""

        def visit_Name(  # pylint: disable=invalid-name,no-self-use,missing-docstring
            self, node: ast.Name
        ) -> ast.expr:
            nonlocal replacements_in_text

            if node.id not in contract.condition_arg_set:
                return node

            assert lambda_inspection is not None
            start, end = lambda_inspection.atok.get_text_range(node)

            if node.id == "_KWARGS":
                replacements_in_text.append(
                    Replace(start=start, end=end, replacement=kwargs_name)
                )
                return ast.Name(id=kwargs_name, ctx=ast.Load())

            replacements_in_text.append(
                Replace(
                    start=start,
                    end=end,
                    replacement="{}[{!r}]".format(kwargs_name, node.id),
                )
            )

            return ast.Subscript(
                value=ast.Name(id=kwargs_name, ctx=ast.Load()),
                slice=ast.Index(value=ast.Constant(value=node.id, kind=None)),
                ctx=ast.Load(),
            )

    lambda_node = ReplaceNameWithSubscriptTransformer().visit(lambda_inspection.node)

    lambda_node.args = ast.arguments(
        posonlyargs=[],
        args=[ast.arg(arg=kwargs_name, annotation=None, type_comment=None)],
        vararg=None,
        kwonlyargs=[],
        kw_defaults=[],
        kwarg=None,
        defaults=[],
    )

    expression_node = ast.Expression(body=lambda_node)

    class AddPositionVisitor(ast.NodeVisitor):
        """Add position to all the nodes if they do not have one already."""

        def generic_visit(self, node: Any) -> None:
            assert lambda_inspection is not None
            node.lineno = lambda_inspection.node.lineno
            node.col_offset = lambda_inspection.node.col_offset

            super().generic_visit(node)

    AddPositionVisitor().visit(expression_node)

    ##
    # Rewrite text
    ##

    if len(replacements_in_text) == 0:
        condition_repr = lambda_inspection.atok.get_text(lambda_inspection.node)
    else:
        # We have to use the text of the whole decorator, not just the lambda since ``atok``
        # refers to the decorator.
        text = lambda_inspection.atok.text

        body_start, body_end = lambda_inspection.atok.get_text_range(
            lambda_inspection.node.body
        )

        parts = []  # type: List[str]

        previous_replace = None  # type: Optional[Replace]
        for replace in replacements_in_text:
            if previous_replace is None:
                parts.append(text[body_start : replace.start])
            else:
                parts.append(text[previous_replace.end : replace.start])

            parts.append(replace.replacement)
            previous_replace = replace

        assert previous_replace is not None
        parts.append(text[previous_replace.end : body_end])
        condition_repr = "lambda {}: {}".format(kwargs_name, "".join(parts))

    ##
    # Recompile the code
    ##

    recompiled_code = compile(
        expression_node,
        filename=f"<rewritten by icontract-hypothesis: {condition_repr}>",
        mode="eval",
    )

    ##
    # Patch the recompiled code such that it uses globals and freevars (closures) from the
    # original condition function
    ##

    # We need to patch all the closured vars in the recompiled code
    # to LOAD_DEREF instead of LOAD_GLOBAL
    closured_vars = {
        name: i for i, name in enumerate(contract.condition.__code__.co_freevars)
    }
    if len(closured_vars) > 256:
        raise NotImplementedError(
            "icontract-hypothesis does not handle condition lambdas with more than 256 closures. "
            "Please create an issue to let us know to implement it: "
            "https://github.com/mristin/icontract-hypothesis/issues"
        )

    # The indices of global vars need to change to correspond to the original condition function.
    global_vars = {
        name: i for i, name in enumerate(contract.condition.__code__.co_names)
    }

    # The first const contains the actual code of the lambda function.
    lambda_code = recompiled_code.co_consts[0]

    opcode_load_global = dis.opmap["LOAD_GLOBAL"]
    opcode_load_deref = dis.opmap["LOAD_DEREF"]
    opcode_extended_arg = dis.opmap["EXTENDED_ARG"]

    bytecode = lambda_code.co_code
    new_bytecode = io.BytesIO()

    # The last instruction must be RETURN_VALUE.
    assert (
        bytecode[-2:]
        == dis.opmap["RETURN_VALUE"].to_bytes(1, byteorder="little") + b"\x00"
    )

    # Since we know that the last instruction is RETURN_VALUE,
    # we can always look up the next opcode.
    for i in range(0, len(bytecode) - 2, 2):
        opcode = bytecode[i]
        oparg = bytecode[i + 1]

        next_opcode = bytecode[i + 2]

        if opcode == opcode_extended_arg and next_opcode == opcode_load_global:
            raise NotImplementedError(
                "icontract-hypothesis does not handle lambdas with more than 256 globals. "
                "Please create an issue to let us know to implement it: "
                "https://github.com/mristin/icontract-hypothesis/issues"
            )

        if opcode == opcode_load_global:
            name = lambda_code.co_names[oparg]
            if name in closured_vars:
                new_bytecode.write(
                    opcode_load_deref.to_bytes(1, byteorder="little")
                    + closured_vars[name].to_bytes(1, byteorder="little")
                )
            elif name in global_vars:
                new_bytecode.write(
                    opcode_load_global.to_bytes(1, byteorder="little")
                    + global_vars[name].to_bytes(1, byteorder="little")
                )
            else:
                raise AssertionError(
                    f"Unexpected case where the opcode was {opcode} ({(dis.opmap[opcode])!r}"
                    f"and the name was {name!r}, but the original closures were {closured_vars!r} "
                    f"and the original globals were {global_vars!r}, respectively."
                )
        else:
            new_bytecode.write(bytecode[i : i + 2])

    new_bytecode.write(bytecode[-2:])

    if sys.version_info < (3, 8):
        new_lambda_code = types.CodeType(
            lambda_code.co_argcount,
            lambda_code.co_kwonlyargcount,
            lambda_code.co_nlocals,
            lambda_code.co_stacksize,
            lambda_code.co_flags,
            new_bytecode.getvalue(),
            lambda_code.co_consts,
            contract.condition.__code__.co_names,
            lambda_code.co_varnames,
            "<recompiled by icontract-hypothesis>",
            "<lambda>",
            lambda_code.co_firstlineno,
            lambda_code.co_lnotab,
            contract.condition.__code__.co_freevars,
            lambda_code.co_cellvars,
        )
    else:
        assert isinstance(lambda_code, types.CodeType)
        new_lambda_code = lambda_code.replace(
            co_code=new_bytecode.getvalue(),
            co_names=contract.condition.__code__.co_names,
            co_filename="<recompiled by icontract-hypothesis>",
            co_name="<lambda>",
            co_freevars=contract.condition.__code__.co_freevars,
        )

    recompiled_function = eval(  # pylint: disable=eval-used
        recompiled_code, dict(globals())
    )

    ##
    # Re-wire the globals and the closure
    ##

    rewired_function = types.FunctionType(
        new_lambda_code,
        contract.condition.__globals__,  # type: ignore
        name="<lambda>",
        argdefs=recompiled_function.__defaults__,
        closure=contract.condition.__closure__,  # type: ignore
    )

    rewired_function.__icontract_hypothesis_source_code__ = condition_repr  # type: ignore

    return rewired_function


def _infer_strategy_from_conjunction(
    type_hints: Mapping[str, Any],
    conjunction: List[icontract._types.Contract],
    self_instance: Optional[Any],
) -> hypothesis.strategies.SearchStrategy[Any]:
    """
    Infer the strategy that satisfies the given conjunction of contracts.

    If ``self_instance`` has been specified, it is included as 'self' in the initial
    fixed dictionary of strategies.
    """
    ##
    # Group single-argument contracts by the argument,
    # keep the list of the zero or multi-argument contracts
    ##

    single_argument_contracts = (
        dict()
    )  # type: MutableMapping[str, List[icontract._types.Contract]]

    non_single_argument_contracts = []  # type: List[icontract._types.Contract]

    for contract in conjunction:
        if len(contract.condition_args) == 1 and contract.condition_args[0] not in (
            "_ARGS",
            "_KWARGS",
        ):
            arg_name = contract.condition_args[0]

            if arg_name not in single_argument_contracts:
                single_argument_contracts[arg_name] = [contract]
            else:
                single_argument_contracts[arg_name].append(contract)
        else:
            non_single_argument_contracts.append(contract)

    if self_instance is not None and "self" in type_hints:
        raise AssertionError(
            f"Unexpected 'self' in ``type_hints`` {type_hints!r} "
            f"while ``self_instance`` is given as {self_instance}. "
            f"We expected 'self' to be removed from type hints in this case."
        )

    mapping = {
        arg_name: _infer_strategy_for_argument(
            arg_name=arg_name,
            type_hint=type_hint,
            contracts=single_argument_contracts.get(arg_name, None),
        )
        for arg_name, type_hint in type_hints.items()
    }

    if self_instance is not None and "self" in mapping:
        raise AssertionError(
            f"Unexpected 'self' in mapping of initial strategies {mapping!r} "
            f"while ``self_instance`` is given as {self_instance}."
        )

    if self_instance is not None:
        # The contracts on ``self`` will be handled in a special way, see below.
        mapping["self"] = hypothesis.strategies.just(self_instance)

    strategy = hypothesis.strategies.fixed_dictionaries(mapping)

    # We need to chain contracts on ``self`` *after* the initial dictionary as Hypothesis
    # optimizes away ``.filter`` on ``just`` strategy in representation (though the filtering
    # is actually performed).
    #
    # See https://github.com/HypothesisWorks/hypothesis/pull/2688 (code) and
    # https://github.com/HypothesisWorks/hypothesis/issues/2036 (issue) and
    # https://github.com/HypothesisWorks/hypothesis/issues/2964 (issue).
    if self_instance is not None and "self" in single_argument_contracts:
        for contract in single_argument_contracts["self"]:
            strategy = strategy.filter(
                _rewrite_condition_as_filter_on_kwargs(contract=contract)
            )

    for contract in non_single_argument_contracts:
        strategy = strategy.filter(
            _rewrite_condition_as_filter_on_kwargs(contract=contract)
        )

    return strategy


def _create_strategy_only_from_type_hints(
    type_hints: Mapping[str, Any], self_instance: Optional[Any]
) -> hypothesis.strategies.SearchStrategy[Any]:
    """
    Create a strategy based only on type hints.

    For example, this is the strategy that you can infer if there are no preconditions.

    If ``self_instance`` is specified, it is included as a single-example strategy for 'self'.
    """
    mapping = {
        arg_name: hypothesis.strategies.from_type(arg_type)
        for arg_name, arg_type in type_hints.items()
    }

    if self_instance is not None and "self" in mapping:
        raise AssertionError(
            f"Unexpected 'self' in strategy mapping {mapping!r} "
            f"while self_instance has been specified: {self_instance}"
        )

    if self_instance is not None:
        mapping["self"] = hypothesis.strategies.just(self_instance)

    return hypothesis.strategies.fixed_dictionaries(mapping)


def infer_strategy(
    func: CallableT,
    localns: Optional[Dict[str, Any]] = None,
    globalns: Optional[Dict[str, Any]] = None,
) -> hypothesis.strategies.SearchStrategy[Any]:
    r"""
    Infer the search strategy of the arguments for the given function.

    You need to supply ``localns`` and ``globalns``, the local namespace and the global namespace
    of the function, respectively, in case there are forward references which can not be
    automatically resolved. For example, if you have nested classes and functions.
    See PEP 563 and the bug reports related to ``typing.get_type_hints`` for more details.

    Apart from the internal usage, this function is mainly meant for manual inspection
    of the inferred strategies.

    Here is an example how you can debug what strategies will be used to test ``some_func``:

    >>> import icontract
    >>> import icontract_hypothesis

    >>> @icontract.require(lambda x: x > 0)
    ... def some_func(x: int) -> None:
    ...    ...

    >>> strategy = icontract_hypothesis.infer_strategy(some_func)
    >>> str(strategy)
    "fixed_dictionaries({'x': integers(min_value=1)})"
    """
    parameters = inspect.signature(func).parameters
    if len(parameters) == 0 and not hasattr(func, "__self__"):
        return hypothesis.strategies.just(dict())

    for name, parameter in parameters.items():
        if parameter.kind == inspect.Parameter.POSITIONAL_ONLY:
            raise NotImplementedError(
                "Inferring strategy of a function with positional-only arguments is not "
                "currently implemented in icontract-hypothesis. "
                "Please create an issue on GitHub if you need this feature: "
                "https://github.com/mristin/icontract-hypothesis/issues"
            )
        else:
            pass

    type_hints = typing.get_type_hints(func, localns=localns, globalns=globalns)
    if "return" in type_hints:
        del type_hints["return"]

    parameter_set = set(parameters.keys())

    # We need to take special care when handling the bound methods whose contracts involve 'self'.
    if hasattr(func, "__self__") and "self" in parameter_set:
        parameter_set.remove("self")

        if "self" in type_hints:
            del type_hints["self"]

    if "self" in parameter_set and func.__name__ == "__init__":
        # The self instance does not exist yet in the __init__ so we can not generate it and
        # can safely remove it from the parameters.
        parameter_set.remove("self")

        if "self" in type_hints:
            del type_hints["self"]

    # If the class specifies its own ``__new__`` we have to remove the first parameter corresponding
    # to the class from the list of parameters. Python will supply it by convention.
    #
    # However, in case of ``object.__new__``, there is no ``cls`` as the first argument!
    #
    elif (
        func.__name__ == "__new__"
        and func != object.__new__  # pylint: disable=comparison-with-callable
    ):
        if len(parameters.keys()) > 0:
            # Remove the first parameter which points to the class.
            #
            # The dunder ``__new__`` is special as we do not want to generate the first argument
            # corresponding to the class.
            cls_parameter = next(k for k in parameters.keys())

            parameter_set.remove(cls_parameter)

            if cls_parameter in type_hints:
                del type_hints[cls_parameter]

    elif (
        not hasattr(func, "__self__")
        and "self" in parameters
        and "self" not in type_hints
        and inspect.isfunction(func)
    ):
        # This might be an unbound instance method.
        # There is no way in Python 3 how we can directly figure out the class of an unbound
        # method, so we try to hack our way through using ``__qualname__``.
        #
        # See https://stackoverflow.com/questions/3589311/get-defining-class-of-unbound-method-object-in-python-3
        if ".<locals>." not in func.__qualname__:
            parts = func.__qualname__.rsplit(".", 1)
            if len(parts) >= 2:
                maybe_cls = getattr(inspect.getmodule(func), parts[0], None)
                if inspect.isclass(maybe_cls):
                    type_hints["self"] = maybe_cls

    typed_args = set(type_hints)

    for name, parameter in parameters.items():
        if parameter.kind in (
            inspect.Parameter.VAR_KEYWORD,
            inspect.Parameter.VAR_POSITIONAL,
        ):
            # Ignore variable keyword and positional arguments in further analysis
            parameter_set.remove(name)

            if name in type_hints:
                typed_args.remove(name)
                del type_hints[name]

    if typed_args != parameter_set:
        raise TypeError(
            (
                "No search strategy could be inferred for the function: {}; "
                "the following arguments are missing the type annotations: {};\n\n"
                "sorted typed_args was {}, sorted parameter_set was {}"
            ).format(
                func,
                list(parameter_set.difference(typed_args)),
                sorted(typed_args),
                sorted(parameter_set),
            )
        )

    checker = icontract._checkers.find_checker(func)

    if checker is None:
        return _create_strategy_only_from_type_hints(
            type_hints=type_hints, self_instance=getattr(func, "__self__", None)
        )

    maybe_preconditions = getattr(checker, "__preconditions__", None)

    preconditions = None  # type: Optional[List[List[icontract._types.Contract]]]
    if maybe_preconditions is not None:
        assert isinstance(maybe_preconditions, list)
        assert all(isinstance(conjunction, list) for conjunction in maybe_preconditions)
        assert all(
            isinstance(contract, icontract._types.Contract)
            for conjunction in maybe_preconditions
            for contract in conjunction
        )

        preconditions = cast(List[List[icontract._types.Contract]], maybe_preconditions)

    if preconditions is None or len(preconditions) == 0:
        return _create_strategy_only_from_type_hints(
            type_hints=type_hints, self_instance=getattr(func, "__self__", None)
        )

    strategies = [
        _infer_strategy_from_conjunction(
            type_hints=type_hints,
            conjunction=conjunction,
            self_instance=getattr(func, "__self__", None),
        )
        for conjunction in preconditions
    ]

    if len(strategies) == 1:
        return strategies[0]

    return hypothesis.strategies.one_of(strategies)


def test_with_inferred_strategy(
    func: CallableT,
    localns: Optional[Dict[str, Any]] = None,
    globalns: Optional[Dict[str, Any]] = None,
) -> None:
    r"""
    Use type hints and contracts to infer the search strategy and test the function.

    You need to supply ``localns`` and ``globalns``, the local namespace and the global namespace
    of the function, respectively, in case there are forward references which can not be
    automatically resolved. For example, if you have nested classes and functions.
    See PEP 563 and the bug reports related to ``typing.get_type_hints`` for more details.

    Here is an example test case which tests a function ``some_func``:

    >>> import unittest
    >>> import icontract
    >>> import icontract_hypothesis

    >>> @icontract.require(lambda x: x > 0)
    ... @icontract.require(lambda x: x < 100)
    ... @icontract.require(lambda y: 0 < y < 100)
    ... def some_func(x: int, y: int) -> None:
    ...     ...

    >>> class TestSomething(unittest.TestCase):
    ...     def test_that_it_works(self) -> None:
    ...         icontract_hypothesis.test_with_inferred_strategy(some_func)

    >>> TestSomething().test_that_it_works()
    """
    pass  # for pydocstyle

    # Since ``self`` is already included in the strategy, we have to unbind the ``func``.
    # Please see ``infer_strategy`` function and how it deals with ``self_instance``.
    func_to_execute = func
    if hasattr(func, "__self__"):
        func_to_execute = func.__func__  # type: ignore

    strategy = infer_strategy(func=func, localns=localns, globalns=globalns)

    def execute(kwargs: Dict[str, Any]) -> None:
        func_to_execute(**kwargs)

    wrapped = hypothesis.given(strategy)(execute)
    wrapped()


def _register_with_hypothesis(cls: Type[T]) -> None:
    """
    Register ``cls`` with Hypothesis based on our custom ``_strategy_for_type``.

    The registration is necessary so that the preconditions on the __init__ are propagated
    in ``hypothesis.strategies.builds``.
    """
    # We should not register abstract classes as this will mislead Hypothesis to instantiate
    # them. However, if a class provides a ``__new__``, we can instantiate it even if it is
    # abstract!
    # pylint: disable=comparison-with-callable
    if inspect.isabstract(cls) and getattr(cls, "__new__") == object.__new__:
        return

    if cls not in hypothesis.strategies._internal.types._global_type_lookup:
        hypothesis.strategies.register_type_strategy(cls, _strategy_for_type(cls))


def _hook_into_icontract_and_hypothesis() -> None:
    """
    Redirect ``icontract._metaclass._register_for_hypothesis``.

    All the classes previously registered by icontract are now re-registered
    by ``_register_with_hypothesis``.
    """
    if not hasattr(icontract._metaclass, "_CONTRACT_CLASSES"):
        return  # already hooked in

    icontract._metaclass._register_for_hypothesis = _register_with_hypothesis

    for cls in icontract._metaclass._CONTRACT_CLASSES:
        _register_with_hypothesis(cls)

    # Delete in order to fail fast
    del icontract._metaclass._CONTRACT_CLASSES

    # Monkey-patch lambda source so that we do not have to introduce
    # strategy classes just for this functionality.
    # See https://github.com/HypothesisWorks/hypothesis/issues/2713
    upstream_extract_lambda_source = (
        hypothesis.internal.reflection.extract_lambda_source
    )
    hypothesis.internal.reflection.extract_lambda_source = lambda f: (
        getattr(f, "__icontract_hypothesis_source_code__", None)
        or upstream_extract_lambda_source(f)  # type: ignore
    )


_hook_into_icontract_and_hypothesis()



# pylint: disable=missing-docstring
# pylint: disable=invalid-name
# pylint: disable=unused-argument
import io
import os
import pathlib
import re
import textwrap
import unittest
import uuid

import hypothesis

from icontract_hypothesis.pyicontract_hypothesis import _general, main, _test


class TestParsingOfParameters(unittest.TestCase):
    def test_subcommand_test(self) -> None:
        # fmt: off
        argv = [
            "test",
            "--path", "some_module.py",
            "--include", "include-something",
            "--exclude", "exclude-something",
            "--setting",
            'suppress_health_check=["too_slow"]', 'verbosity="verbose"',
        ]
        # fmt: on

        parser = main._make_argument_parser()
        args, out, err = main._parse_args(parser=parser, argv=argv)
        assert args is not None, "Failed to parse argv {!r}: {}".format(argv, err)

        general, errs = _general.parse_params(args=args)

        self.assertListEqual([], errs)
        assert general is not None
        self.assertListEqual(
            [re.compile(pattern) for pattern in ["include-something"]], general.include
        )
        self.assertListEqual(
            [re.compile(pattern) for pattern in ["exclude-something"]], general.exclude
        )

        test, errs = _test.parse_params(args=args)

        self.assertListEqual([], errs)
        assert test is not None
        self.assertEqual(pathlib.Path("some_module.py"), test.path)

        assert test.settings_parsing is not None

        self.assertListEqual(
            [hypothesis.HealthCheck.too_slow],
            test.settings_parsing.product.__dict__["suppress_health_check"],
        )

        self.assertEqual(
            hypothesis.Verbosity.verbose,
            test.settings_parsing.product.__dict__["verbosity"],
        )

    def test_unknown_settings(self) -> None:
        # fmt: off
        argv = [
            "test",
            "--path", "some_module.py",
            "--setting", "totally_invalid=[2, 3]",
        ]
        # fmt: on

        parser = main._make_argument_parser()
        args, out, err = main._parse_args(parser=parser, argv=argv)
        assert args is not None, "Failed to parse argv {!r}: {}".format(argv, err)

        general, errs = _general.parse_params(args=args)
        self.assertListEqual([], errs)

        test, errs = _test.parse_params(args=args)
        self.assertListEqual(["Invalid Hypothesis setting: 'totally_invalid'"], errs)

    def test_invalid_settings(self) -> None:
        # fmt: off
        argv = [
            "test",
            "--path", "some_module.py",
            "--setting", "max_examples=-1",
        ]
        # fmt: on

        parser = main._make_argument_parser()
        args, out, err = main._parse_args(parser=parser, argv=argv)
        assert args is not None, "Failed to parse argv {!r}: {}".format(argv, err)

        general, errs = _general.parse_params(args=args)
        self.assertListEqual([], errs)

        test, errs = _test.parse_params(args=args)
        self.assertListEqual(
            [
                "Invalid Hypothesis settings: max_examples=-1 should be at least one. You can "
                "disable example generation with the `phases` setting instead."
            ],
            errs,
        )


class TestTest(unittest.TestCase):
    def test_default_behavior(self) -> None:
        this_dir = pathlib.Path(os.path.realpath(__file__)).parent
        path = (
            this_dir.parent.parent
            / "test_samples/pyicontract_hypothesis/sample_module.py"
        )

        mod, errors = _general.load_module_from_source_file(path=path)
        self.assertListEqual([], errors)
        assert mod is not None

        points, errors = _general.select_function_points(
            source_code=path.read_text(), mod=mod, include=[], exclude=[]
        )
        self.assertListEqual([], errors)

        for point in points:
            test_errors = _test._test_function_point(
                point=point, hypothesis_settings=None
            )
            self.assertListEqual([], test_errors)

        some_func_calls = getattr(mod, "SOME_FUNC_CALLS")
        self.assertEqual(100, some_func_calls)

        another_func_calls = getattr(mod, "ANOTHER_FUNC_CALLS")
        self.assertEqual(100, another_func_calls)

    def test_settings(self) -> None:
        settings_parsing, errors = _test._parse_hypothesis_settings(
            {"max_examples": 10}
        )
        self.assertListEqual([], errors)
        assert settings_parsing is not None

        this_dir = pathlib.Path(os.path.realpath(__file__)).parent
        path = (
            this_dir.parent.parent
            / "test_samples/pyicontract_hypothesis/sample_module.py"
        )

        mod, errors = _general.load_module_from_source_file(path=path)
        self.assertListEqual([], errors)
        assert mod is not None

        points, errors = _general.select_function_points(
            source_code=path.read_text(), mod=mod, include=[], exclude=[]
        )
        self.assertListEqual([], errors)

        for point in points:
            test_errors = _test._test_function_point(
                point=point, hypothesis_settings=settings_parsing.product
            )
            self.assertListEqual([], test_errors)

        some_func_calls = getattr(mod, "SOME_FUNC_CALLS")
        self.assertEqual(10, some_func_calls)

        another_func_calls = getattr(mod, "ANOTHER_FUNC_CALLS")
        self.assertEqual(10, another_func_calls)


class TestTestViaSmoke(unittest.TestCase):
    """Perform smoke testing of the "test" command."""

    def test_nonexisting_file(self) -> None:
        path = "doesnt-exist.{}".format(uuid.uuid4())
        argv = ["test", "--path", path]

        stdout = io.StringIO()
        stderr = io.StringIO()

        exit_code = main.run(argv=argv, stdout=stdout, stderr=stderr)

        self.assertEqual(
            "The file to be tested does not exist: {}".format(path),
            stderr.getvalue().strip(),
        )
        self.assertEqual(exit_code, 1)

    def test_common_case(self) -> None:
        this_dir = pathlib.Path(os.path.realpath(__file__)).parent
        pth = (
            this_dir.parent.parent
            / "test_samples/pyicontract_hypothesis/sample_module.py"
        )

        argv = ["test", "--path", str(pth)]

        stdout = io.StringIO()
        stderr = io.StringIO()

        # This is merely a smoke test.
        exit_code = main.run(argv=argv, stdout=stdout, stderr=stderr)

        self.assertEqual("", stderr.getvalue())
        self.assertEqual(exit_code, 0)

        out = re.sub("\(time delta [^)]*\)", "(time delta <erased>)", stdout.getvalue())

        self.assertEqual(
            textwrap.dedent(
                """\
            Tested some_func at line 10 (time delta <erased>).
            Tested another_func at line 19 (time delta <erased>).
            Tested yet_another_func at line 28 (time delta <erased>).
            """
            ),
            out,
        )

    def test_with_settings(self) -> None:
        this_dir = pathlib.Path(os.path.realpath(__file__)).parent
        pth = (
            this_dir.parent.parent
            / "test_samples/pyicontract_hypothesis/sample_module.py"
        )

        argv = ["test", "--path", str(pth), "--settings", "max_examples=5"]

        stdout = io.StringIO()
        stderr = io.StringIO()

        # This is merely a smoke test.
        exit_code = main.run(argv=argv, stdout=stdout, stderr=stderr)

        self.assertEqual("", stderr.getvalue())
        self.assertEqual(exit_code, 0)

    def test_with_include_exclude(self) -> None:
        this_dir = pathlib.Path(os.path.realpath(__file__)).parent
        pth = (
            this_dir.parent.parent
            / "test_samples/pyicontract_hypothesis/sample_module.py"
        )

        # fmt: off
        argv = [
            "test",
            "--path", str(pth),
            "--include", ".*_func",
            "--exclude", "some.*",
        ]
        # fmt: on

        stdout = io.StringIO()
        stderr = io.StringIO()

        # This is merely a smoke test.
        exit_code = main.run(argv=argv, stdout=stdout, stderr=stderr)

        self.assertEqual("", stderr.getvalue())
        self.assertEqual(exit_code, 0)


class TestInspect(unittest.TestCase):
    """Perform smoke testing of the "test" command."""

    def test_common_case(self) -> None:
        this_dir = pathlib.Path(os.path.realpath(__file__)).parent
        pth = (
            this_dir.parent.parent
            / "test_samples/pyicontract_hypothesis/sample_module.py"
        )

        argv = ["test", "--inspect", "--path", str(pth)]

        stdout = io.StringIO()
        stderr = io.StringIO()

        # This is merely a smoke test.
        exit_code = main.run(argv=argv, stdout=stdout, stderr=stderr)

        self.assertEqual("", stderr.getvalue())
        self.assertEqual(exit_code, 0)

        self.assertEqual(
            textwrap.dedent(
                """\
            some_func at line 10:
               hypothesis.given(
                   fixed_dictionaries({'x': integers(min_value=1)})
               )
            
            another_func at line 19:
               hypothesis.given(
                   fixed_dictionaries({'x': integers(min_value=1).filter(lambda x: square_greater_than_zero(x))})
               )
            
            yet_another_func at line 28:
               hypothesis.given(
                   fixed_dictionaries({'x': integers(), 'y': integers()}).filter(lambda d: d['x'] < d['y'])
               )
            """
            ),
            stdout.getvalue(),
        )

    def test_with_settings(self) -> None:
        this_dir = pathlib.Path(os.path.realpath(__file__)).parent
        pth = (
            this_dir.parent.parent
            / "test_samples/pyicontract_hypothesis/sample_module.py"
        )

        # fmt: off
        argv = [
            "test",
            "--inspect",
            "--path", str(pth),
            "--settings",
            "max_examples=5",
            'suppress_health_check=["too_slow"]']
        # fmt: on

        stdout = io.StringIO()
        stderr = io.StringIO()

        # This is merely a smoke test.
        exit_code = main.run(argv=argv, stdout=stdout, stderr=stderr)

        self.assertEqual("", stderr.getvalue())
        self.assertEqual(exit_code, 0)

        self.assertEqual(
            textwrap.dedent(
                """\
                some_func at line 10:
                   hypothesis.given(
                       fixed_dictionaries({'x': integers(min_value=1)})
                   )
                   hypothesis.settings(
                       max_examples=5,
                       suppress_health_check=[HealthCheck.too_slow]
                   )
                
                another_func at line 19:
                   hypothesis.given(
                       fixed_dictionaries({'x': integers(min_value=1).filter(lambda x: square_greater_than_zero(x))})
                   )
                   hypothesis.settings(
                       max_examples=5,
                       suppress_health_check=[HealthCheck.too_slow]
                   )
                
                yet_another_func at line 28:
                   hypothesis.given(
                       fixed_dictionaries({'x': integers(), 'y': integers()}).filter(lambda d: d['x'] < d['y'])
                   )
                   hypothesis.settings(
                       max_examples=5,
                       suppress_health_check=[HealthCheck.too_slow]
                   )
            """
            ),
            stdout.getvalue(),
        )


if __name__ == "__main__":
    unittest.main()



# pylint: disable=missing-docstring
# pylint: disable=invalid-name
# pylint: disable=unused-argument
import os
import pathlib
import re
import unittest
from typing import Pattern

import icontract_hypothesis.pyicontract_hypothesis._general as _general


class TestLineRangeRe(unittest.TestCase):
    def test_only_first(self) -> None:
        mtch = _general._LINE_RANGE_RE.match(" 123 ")
        assert mtch is not None

        self.assertEqual("123", mtch.group("first"))
        self.assertIsNone(
            mtch.group("last"), "Unexpected last group: {}".format(mtch.group("last"))
        )

    def test_first_and_last(self) -> None:
        mtch = _general._LINE_RANGE_RE.match(" 123 - 435 ")
        assert mtch is not None

        self.assertEqual("123", mtch.group("first"))
        self.assertEqual("435", mtch.group("last"))

    def test_no_match(self) -> None:
        mtch = _general._LINE_RANGE_RE.match("123aa")
        assert mtch is None, "Expected no match, but got: {}".format(mtch)


class TestParsingOfPointSpecs(unittest.TestCase):
    def test_single_line(self) -> None:
        text = "123"
        point_spec, errors = _general._parse_point_spec(text=text)

        self.assertListEqual([], errors)
        assert isinstance(point_spec, _general._LineRange)
        self.assertEqual(123, point_spec.first)
        self.assertEqual(123, point_spec.last)

    def test_line_range(self) -> None:
        text = "123-345"
        point_spec, errors = _general._parse_point_spec(text=text)

        self.assertListEqual([], errors)
        assert isinstance(point_spec, _general._LineRange)
        self.assertEqual(123, point_spec.first)
        self.assertEqual(345, point_spec.last)

    def test_invalid_line_range(self) -> None:
        text = "345-123"
        point_spec, errors = _general._parse_point_spec(text=text)

        assert point_spec is None
        self.assertListEqual(["Unexpected line range (last < first): 345-123"], errors)

    def test_pattern(self) -> None:
        text = r"^do_.*$"
        point_spec, errors = _general._parse_point_spec(text=text)

        self.assertListEqual([], errors)
        assert isinstance(point_spec, Pattern)
        self.assertEqual(text, point_spec.pattern)


class TestSelectFunctionPoints(unittest.TestCase):
    def test_invalid_module(self) -> None:
        this_dir = pathlib.Path(os.path.realpath(__file__)).parent
        path = (
            this_dir.parent.parent
            / "test_samples/pyicontract_hypothesis/sample_invalid_module.py"
        )

        mod, errors = _general.load_module_from_source_file(path=path)
        self.assertListEqual([], errors)
        assert mod is not None

        points, errors = _general.select_function_points(
            source_code=path.read_text(), mod=mod, include=[], exclude=[]
        )

        self.assertListEqual(
            [
                "Unexpected directive on line 8. "
                "Expected '# pyicontract-hypothesis: (disable|enable)', "
                "but got: # pyicontract-hypothesis: disable-once"
            ],
            errors,
        )

    def test_no_include_and_no_exclude(self) -> None:
        this_dir = pathlib.Path(os.path.realpath(__file__)).parent
        path = (
            this_dir.parent.parent
            / "test_samples/pyicontract_hypothesis/sample_module.py"
        )

        mod, errors = _general.load_module_from_source_file(path=path)
        self.assertListEqual([], errors)
        assert mod is not None

        points, errors = _general.select_function_points(
            source_code=path.read_text(), mod=mod, include=[], exclude=[]
        )
        self.assertListEqual([], errors)

        self.assertListEqual(
            ["some_func", "another_func", "yet_another_func"],
            [point.func.__name__ for point in points],
        )

    def test_include_line_range(self) -> None:
        this_dir = pathlib.Path(os.path.realpath(__file__)).parent
        path = (
            this_dir.parent.parent
            / "test_samples/pyicontract_hypothesis/sample_module.py"
        )

        mod, errors = _general.load_module_from_source_file(path=path)
        self.assertListEqual([], errors)
        assert mod is not None

        points, errors = _general.select_function_points(
            source_code=path.read_text(),
            mod=mod,
            # A single line that overlaps the function should be enough to include it.
            include=[_general._LineRange(first=13, last=13)],
            exclude=[],
        )
        self.assertListEqual([], errors)

        self.assertListEqual(["some_func"], [point.func.__name__ for point in points])

    def test_include_pattern(self) -> None:
        this_dir = pathlib.Path(os.path.realpath(__file__)).parent
        path = (
            this_dir.parent.parent
            / "test_samples/pyicontract_hypothesis/sample_module.py"
        )

        mod, errors = _general.load_module_from_source_file(path=path)
        self.assertListEqual([], errors)
        assert mod is not None

        points, errors = _general.select_function_points(
            source_code=path.read_text(),
            mod=mod,
            include=[re.compile(r"^some_.*$")],
            exclude=[],
        )
        self.assertListEqual([], errors)

        self.assertListEqual(["some_func"], [point.func.__name__ for point in points])

    def test_exclude_line_range(self) -> None:
        this_dir = pathlib.Path(os.path.realpath(__file__)).parent
        path = (
            this_dir.parent.parent
            / "test_samples/pyicontract_hypothesis/sample_module.py"
        )

        mod, errors = _general.load_module_from_source_file(path=path)
        self.assertListEqual([], errors)
        assert mod is not None

        points, errors = _general.select_function_points(
            source_code=path.read_text(),
            mod=mod,
            include=[],
            # A single line that overlaps the function should be enough to exclude it.
            exclude=[_general._LineRange(first=13, last=13)],
        )
        self.assertListEqual([], errors)

        self.assertListEqual(
            ["another_func", "yet_another_func"],
            [point.func.__name__ for point in points],
        )

    def test_exclude_pattern(self) -> None:
        this_dir = pathlib.Path(os.path.realpath(__file__)).parent
        path = (
            this_dir.parent.parent
            / "test_samples/pyicontract_hypothesis/sample_module.py"
        )

        mod, errors = _general.load_module_from_source_file(path=path)
        self.assertListEqual([], errors)
        assert mod is not None

        points, errors = _general.select_function_points(
            source_code=path.read_text(),
            mod=mod,
            include=[],
            exclude=[re.compile(r"^some_.*$")],
        )
        self.assertListEqual([], errors)

        self.assertListEqual(
            ["another_func", "yet_another_func"],
            [point.func.__name__ for point in points],
        )


if __name__ == "__main__":
    unittest.main()



"""Automatically test a Python file."""
import argparse
import collections
import datetime
import inspect
import json
import pathlib
import re
import textwrap
from typing import (
    Mapping,
    Any,
    Tuple,
    Optional,
    List,
    MutableMapping,
    Dict,
    TextIO,
)

import hypothesis
import hypothesis._settings
import hypothesis.errors

import icontract_hypothesis
import icontract_hypothesis.pyicontract_hypothesis._general as _general


class Params:
    """Represent parameters of the command "test"."""

    def __init__(
        self,
        path: pathlib.Path,
        settings_parsing: Optional["SettingsParsing"],
        do_inspect: bool,
    ) -> None:
        """Initialize with the given values."""
        self.path = path
        self.settings_parsing = settings_parsing
        self.inspect = do_inspect


_SETTING_STATEMENT_RE = re.compile(
    r"^(?P<identifier>[a-zA-Z_][a-zA-Z_0-9]*)\s*=\s*(?P<value>.*)\s*$"
)

_HYPOTHESIS_SETTINGS_PARAMETERS = inspect.signature(
    hypothesis.settings.__init__
).parameters
_HYPOTHESIS_HEALTH_CHECK_MAP = {
    health_check.name: health_check for health_check in hypothesis.HealthCheck
}
_HYPOTHESIS_VERBOSITY_MAP = {
    verbosity.name: verbosity for verbosity in hypothesis.Verbosity
}
_HYPOTHESIS_PHASE_MAP = {phase.name: phase for phase in hypothesis.Phase}


class SettingsParsing:
    """Represent hypothesis.settings such that it can be inspected down the line."""

    def __init__(self, parsed_kwargs: Mapping[str, Any], product: hypothesis.settings):
        """Initialize with the given values."""
        self.parsed_kwargs = parsed_kwargs
        self.product = product


def _parse_hypothesis_settings(
    settings: Mapping[str, Any]
) -> Tuple[Optional[SettingsParsing], List[str]]:
    """
    Try to parse the setting as Hypothesis settings.

    Return (settings, errors if any).
    """
    errors = []

    kwargs = collections.OrderedDict()  # type: MutableMapping[str, Any]

    for identifier, value in settings.items():
        if identifier not in _HYPOTHESIS_SETTINGS_PARAMETERS:
            errors.append(f"Invalid Hypothesis setting: {identifier!r}")
        else:
            if identifier == "verbosity":
                if value not in _HYPOTHESIS_VERBOSITY_MAP:
                    errors.append(
                        f"Invalid Hypothesis setting {identifier!r}: {value!r}"
                    )
                else:
                    kwargs[identifier] = _HYPOTHESIS_VERBOSITY_MAP[value]

            elif identifier == "phase":
                if not isinstance(value, list):
                    errors.append(
                        f"Invalid Hypothesis setting {identifier!r}: "
                        f"expected a list, but got {value!r}"
                    )
                else:
                    parsed_items = []  # type: List[Any]
                    for item in value:
                        if item not in _HYPOTHESIS_PHASE_MAP:
                            errors.append(
                                f"Invalid {hypothesis.Phase.__name__} in the Hypothesis setting "
                                f"{identifier!r}: {item!r}"
                            )
                        else:
                            parsed_items.append(_HYPOTHESIS_PHASE_MAP[item])
                    kwargs[identifier] = parsed_items

            elif identifier == "suppress_health_check":
                if not isinstance(value, list):
                    errors.append(
                        f"Invalid Hypothesis setting {identifier!r}: "
                        f"expected a list, but got {value!r}"
                    )
                else:
                    parsed_items = []
                    for item in value:
                        if item not in _HYPOTHESIS_HEALTH_CHECK_MAP:
                            errors.append(
                                f"Invalid {hypothesis.HealthCheck.__name__} in the setting "
                                f"{identifier!r}: {item!r}"
                            )
                        else:
                            parsed_items.append(_HYPOTHESIS_HEALTH_CHECK_MAP[item])
                    kwargs[identifier] = parsed_items
            else:
                kwargs[identifier] = value

    if errors:
        return None, errors

    try:
        return (
            SettingsParsing(
                parsed_kwargs=kwargs, product=hypothesis.settings(**kwargs)
            ),
            [],
        )
    except hypothesis.errors.InvalidArgument as error:
        return None, [f"Invalid Hypothesis settings: {error}"]


def parse_params(args: argparse.Namespace) -> Tuple[Optional[Params], List[str]]:
    """
    Try to parse the parameters of the command "test".

    Return (parsed parameters, errors if any).
    """
    errors = []  # type: List[str]

    path = pathlib.Path(args.path)

    settings = collections.OrderedDict()  # type: MutableMapping[str, Any]

    settings_parsing = None  # type: Optional[SettingsParsing]

    if args.settings is not None:
        for i, statement in enumerate(args.settings):
            mtch = _SETTING_STATEMENT_RE.match(statement)
            if not mtch:
                errors.append(
                    "Invalid setting statement {}. Expected statement to match {}, "
                    "but got: {}".format(
                        i + 1, _SETTING_STATEMENT_RE.pattern, statement
                    )
                )

                return None, errors

            identifier = mtch.group("identifier")
            value_str = mtch.group("value")

            try:
                value = json.loads(value_str)
            except json.decoder.JSONDecodeError as error:
                errors.append(
                    "Failed to parse the value of the setting {}: {}".format(
                        identifier, error
                    )
                )
                return None, errors

            settings[identifier] = value

        settings_parsing, settings_errors = _parse_hypothesis_settings(
            settings=settings
        )
        if settings_errors:
            errors.extend(settings_errors)
        else:
            assert settings_parsing is not None

    if errors:
        return None, errors

    return (
        Params(path=path, settings_parsing=settings_parsing, do_inspect=args.inspect),
        errors,
    )


def _test_function_point(
    point: _general.FunctionPoint, hypothesis_settings: Optional[hypothesis.settings]
) -> List[str]:
    """
    Test a single function point.

    Return errors if any.
    """
    errors = []  # type: List[str]

    func = point.func  # Optimize the look-up

    try:
        strategy = icontract_hypothesis.infer_strategy(func=func)
    except AssertionError:
        raise
    except Exception as error:
        error_as_str = str(error)
        if error_as_str == "":
            error_as_str = str(type(error))

        errors.append(
            f"Inference of the search strategy failed for the function: {func}. "
            f"The error was: {error_as_str}"
        )
        return errors

    def execute(kwargs: Dict[str, Any]) -> None:
        func(**kwargs)

    wrapped = hypothesis.given(strategy)(execute)
    if hypothesis_settings:
        wrapped = hypothesis_settings(wrapped)

    try:
        wrapped()
    except hypothesis.errors.FailedHealthCheck as err:
        return [
            f"Failed to test the function: "
            f"{func.__name__} from {inspect.getfile(func)} "
            f"due to the failed Hypothesis health check: {err}"
        ]

    return []


def _inspect_test_function_point(
    point: _general.FunctionPoint, settings_parsing: Optional[SettingsParsing]
) -> Tuple[str, List[str]]:
    """
    Inspect what test will executed to test a single function point.

    Return (inspection, errors if any).
    """
    errors = []  # type: List[str]

    func = point.func  # Optimize the look-up

    try:
        strategy = icontract_hypothesis.infer_strategy(func=func)
    except Exception as error:
        errors.append(
            (
                "Inference of the search strategy failed for the function: {}. "
                "The error was: {}"
            ).format(func, error)
        )
        return "", errors

    parts = [
        textwrap.dedent(
            """\
        hypothesis.given(
        {}
        )"""
        ).format(textwrap.indent(str(strategy), "    "))
    ]

    if settings_parsing and len(settings_parsing.parsed_kwargs) > 0:
        if len(settings_parsing.parsed_kwargs) == 1:
            identifier, value = list(settings_parsing.parsed_kwargs.items())[0]

            settings_str = f"hypothesis.settings({identifier}={value})"
        else:
            lines = ["hypothesis.settings("]
            items = list(settings_parsing.parsed_kwargs.items())
            for i, (identifier, value) in enumerate(items):
                if i < len(items) - 1:
                    lines.append(f"    {identifier}={value},")
                else:
                    lines.append(f"    {identifier}={value}")
            lines.append(")")

            settings_str = "\n".join(lines)

        parts.append(settings_str)

    return "\n".join(parts), []


def test(general: _general.Params, command: Params, stdout: TextIO) -> List[str]:
    """
    Test the specified functions.

    Return errors if any.
    """
    if not command.path.exists():
        return ["The file to be tested does not exist: {}".format(command.path)]

    try:
        source_code = command.path.read_text(encoding="utf-8")
    except Exception as error:
        return ["Failed to read the file {}: {}".format(command.path, error)]

    mod, errors = _general.load_module_from_source_file(path=command.path)
    if errors:
        return errors

    assert mod is not None

    points, errors = _general.select_function_points(
        source_code=source_code,
        mod=mod,
        include=general.include,
        exclude=general.exclude,
    )
    if errors:
        return errors

    if command.inspect:
        printed_previously = False
        for point in points:
            inspection, test_errors = _inspect_test_function_point(
                point=point, settings_parsing=command.settings_parsing
            )
            errors.extend(test_errors)

            if not test_errors:
                if printed_previously:
                    stdout.write("\n")

                stdout.write(f"{point.func.__name__} at line {point.first_row}:\n")
                stdout.write(textwrap.indent(inspection, "   "))
                stdout.write("\n")
                printed_previously = True
    else:
        for point in points:
            start = datetime.datetime.now()
            test_errors = _test_function_point(
                point=point,
                hypothesis_settings=(
                    command.settings_parsing.product
                    if command.settings_parsing is not None
                    else None
                ),
            )
            errors.extend(test_errors)

            duration = datetime.datetime.now() - start

            if not test_errors:
                stdout.write(
                    f"Tested {point.func.__name__} at line {point.first_row} "
                    f"(time delta {duration}).\n"
                )

    if errors:
        return errors

    return []



import io
import unittest

from icontract_hypothesis.pyicontract_hypothesis import main


class TestParsingOfParameters(unittest.TestCase):
    def test_no_command(self) -> None:
        argv = ["-m", "some_module"]

        stdout, stderr = io.StringIO(), io.StringIO()

        main.run(argv=argv, stdout=stdout, stderr=stderr)

        self.assertEqual("", stdout.getvalue())
        self.assertEqual(
            """\
usage: pyicontract-hypothesis [-h] {test,ghostwrite} ...
pyicontract-hypothesis: error: argument command: invalid choice: 'some_module' (choose from 'test', 'ghostwrite')
""",
            stderr.getvalue(),
        )


if __name__ == "__main__":
    unittest.main()



"""Provide an invalid module for testing pyicontract-hypothesis."""

import icontract


@icontract.require(lambda x: x > 0)
def some_func(x: int) -> None:
    # pyicontract-hypothesis: disable-once
    pass



"""Test the automatic testing with inferred strategies."""

# pylint: disable=missing-docstring
# pylint: disable=invalid-name
# pylint: disable=unused-argument
# pylint: disable=no-value-for-parameter
import abc
import dataclasses
import enum
import math
import re
import sys
import unittest
from typing import List, NamedTuple, Union, Optional, Any, Mapping, Sequence

if sys.version_info >= (3, 8):
    from typing import TypedDict

import hypothesis
import hypothesis.strategies as st
import hypothesis.errors
import icontract

import icontract_hypothesis

SOME_GLOBAL_CONST = 0


# noinspection PyUnusedLocal,PyPep8Naming
class TestWithInferredStrategies(unittest.TestCase):
    def test_fail_without_type_hints(self) -> None:
        @icontract.require(lambda x: x > 0)
        def some_func(x) -> None:  # type: ignore
            pass

        type_error = None  # type: Optional[TypeError]
        try:
            icontract_hypothesis.test_with_inferred_strategy(some_func)
        except TypeError as err:
            type_error = err

        assert type_error is not None
        self.assertTrue(
            re.match(
                r"No search strategy could be inferred for the function: <function .*>; "
                r"the following arguments are missing the type annotations: \['x']",
                str(type_error),
            ),
            str(type_error),
        )

    def test_without_preconditions(self) -> None:
        def some_func(x: int) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual("fixed_dictionaries({'x': integers()})", str(strategy))

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_unmatched_pattern(self) -> None:
        @icontract.require(lambda x: x > 0 and x > math.sqrt(x))
        def some_func(x: float) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            "fixed_dictionaries({'x': floats().filter(lambda x: x > 0 and x > math.sqrt(x))})",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_resorting_to_from_type(self) -> None:
        # We can not handle ``.startswith`` at the moment, so we expect
        # ``from_type`` Hypothesis strategy followed by a filter.
        @icontract.require(lambda x: x.startswith("something"))
        def some_func(x: str) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            "fixed_dictionaries({'x': text().filter(lambda x: x.startswith(\"something\"))})",
            str(strategy),
        )

        # We explicitly do not test with this strategy as it will not pass the health check.

    def test_snippet_given(self) -> None:
        @icontract.require(lambda x, y: x < y)
        def some_func(x: float, y: float) -> None:
            pass

        @hypothesis.given(
            st.fixed_dictionaries({"x": st.floats(), "y": st.floats()}).filter(
                lambda d: d["x"] < d["y"]
            )
        )
        def test(kwargs: Mapping[str, Any]) -> None:
            some_func(**kwargs)

        test()

    def test_multi_argument_contract(self) -> None:
        @icontract.require(lambda x, y: x < y)
        def some_func(x: float, y: float) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            "fixed_dictionaries({'x': floats(), 'y': floats()}).filter(lambda d: d['x'] < d['y'])",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_multi_argument_contract_with_closure(self) -> None:
        SOME_CONSTANT = -math.inf
        ANOTHER_CONSTANT = math.inf

        # We must have -512 and +512 so that EXTENDED_ARG opcode is tested as well.
        @icontract.require(
            lambda x, y: SOME_CONSTANT
            < x - 512
            < SOME_GLOBAL_CONST
            < y + 512
            < ANOTHER_CONSTANT
        )
        def some_func(x: float, y: float) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            """\
fixed_dictionaries({'x': floats(), 'y': floats()}).filter(lambda d: SOME_CONSTANT
    < d['x'] - 512
    < SOME_GLOBAL_CONST
    < d['y'] + 512
    < ANOTHER_CONSTANT)""",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_condition_with_function(self) -> None:
        def some_precondition(x: int) -> bool:
            return x > 0

        @icontract.require(some_precondition)
        def some_func(x: int) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            "fixed_dictionaries({'x': integers().filter(some_precondition)})",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_condition_with_function_on__KWARGS(self) -> None:
        def some_precondition(_KWARGS: Mapping[str, Any]) -> bool:
            return len(_KWARGS) == 1

        @icontract.require(some_precondition)
        def some_func(x: int) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            "fixed_dictionaries({'x': integers()}).filter(lambda d: some_precondition(_KWARGS=d))",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_condition_with_function_on_argument_and_KWARGS(self) -> None:
        def some_precondition(x: int, _KWARGS: Mapping[str, Any]) -> bool:
            return x > len(_KWARGS)

        @icontract.require(some_precondition)
        def some_func(x: int, **kwargs: Any) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            "fixed_dictionaries({'x': integers()})"
            ".filter(lambda d: some_precondition(x=d['x'], _KWARGS=d))",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_condition_on_kwargs(self) -> None:
        @icontract.require(lambda _KWARGS: len(_KWARGS) == 1)
        def some_func(x: int) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            "fixed_dictionaries({'x': integers()}).filter(lambda d: len(d) == 1)",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_condition_without_arguments(self) -> None:
        @icontract.require(lambda: SOME_GLOBAL_CONST >= 0)
        def some_func(x: int) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            "fixed_dictionaries({'x': integers()}).filter(lambda: SOME_GLOBAL_CONST >= 0)",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_initial_kwargs_conflicts_with_local_but_not_with_any_global_or_closure(
        self,
    ) -> None:
        @icontract.require(lambda d, e: d < e)
        def some_func(d: int, e: int) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            "fixed_dictionaries({'d': integers(), 'e': integers()})"
            ".filter(lambda d: d['d'] < d['e'])",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_initial_kwargs_conflicts_with_the_closure(self) -> None:
        d = 0

        @icontract.require(lambda x: x + d >= 0)
        @icontract.require(lambda x, y: math.sqrt(x + d) < y)
        def some_func(x: int, y: int) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            "fixed_dictionaries("
            "{'x': integers().filter(lambda x: x + d >= 0), "
            "'y': integers()})"
            ".filter(lambda _d: math.sqrt(_d['x'] + d) < _d['y'])",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(some_func)


# noinspection PyUnusedLocal
class TestWithInferredStrategiesOnClasses(unittest.TestCase):
    def test_no_preconditions_and_no_argument_init(self) -> None:
        class A:
            def __repr__(self) -> str:
                return "A()"

        def some_func(a: A) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual("fixed_dictionaries({'a': builds(A)})", str(strategy))

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_no_preconditions_and_init(self) -> None:
        class A:
            def __init__(self, x: int):
                self.x = x

            def __repr__(self) -> str:
                return "A(x={})".format(self.x)

        def some_func(a: A) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual("fixed_dictionaries({'a': builds(A)})", str(strategy))

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_map_snippet(self) -> None:
        class A(icontract.DBC):
            def __init__(self, x: int, y: int):
                self.x = x
                self.y = y

        def some_func(a: A) -> None:
            pass

        @hypothesis.given(
            a=st.fixed_dictionaries({"x": st.integers(), "y": st.integers()})
            .filter(lambda d: d["x"] < d["y"])
            .map(lambda d: A(**d))
        )
        def test(a: A) -> None:
            some_func(a)

        test()

    def test_from_type(self) -> None:
        class A(icontract.DBC):
            @icontract.require(lambda x, y: x < y)
            def __init__(self, x: int, y: int):
                self.x = x
                self.y = y

            def __repr__(self) -> str:
                return "A(x={}, y={})".format(self.x, self.y)

        strategy = hypothesis.strategies.from_type(A)
        self.assertEqual(
            "fixed_dictionaries("
            "{'x': integers(), 'y': integers()})"
            ".filter(lambda d: d['x'] < d['y'])"
            ".map(lambda d: A(**d))",
            str(strategy),
        )

        @hypothesis.given(a=hypothesis.strategies.from_type(A))
        def test(a: A) -> None:
            ...

        test()

    def test_preconditions_with_multiple_arguments(self) -> None:
        class A(icontract.DBC):
            @icontract.require(lambda x, y: x < y)
            def __init__(self, x: int, y: int):
                self.x = x
                self.y = y

            def __repr__(self) -> str:
                return "A(x={}, y={})".format(self.x, self.y)

        def some_func(a: A) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            "fixed_dictionaries({'a': fixed_dictionaries({'x': integers(), 'y': integers()})"
            ".filter(lambda d: d['x'] < d['y'])"
            ".map(lambda d: A(**d))})",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_preconditions_with_heuristics(self) -> None:
        class A(icontract.DBC):
            @icontract.require(lambda x: x > 0)
            def __init__(self, x: int):
                self.x = x

            def __repr__(self) -> str:
                return "A(x={})".format(self.x)

        def some_func(a: A) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            "fixed_dictionaries("
            "{'a': fixed_dictionaries({'x': integers(min_value=1)})"
            ".map(lambda d: A(**d))})",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_preconditions_without_heuristics(self) -> None:
        class A(icontract.DBC):
            @icontract.require(lambda x: x > 0)
            @icontract.require(lambda x: x > math.sqrt(x))
            def __init__(self, x: float):
                self.x = x

            def __repr__(self) -> str:
                return "A(x={})".format(self.x)

        def some_func(a: A) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            "fixed_dictionaries("
            "{'a': fixed_dictionaries("
            "{'x': floats(min_value=0, exclude_min=True)"
            ".filter(lambda x: x > math.sqrt(x))})"
            ".map(lambda d: A(**d))})",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_with_weakened_preconditions(self) -> None:
        class A(icontract.DBC):
            @icontract.require(lambda x: 0 < x < 20)
            @icontract.require(lambda x: x % 3 == 0)
            def some_func(self, x: int) -> None:
                pass

        class B(A):
            @icontract.require(lambda x: 0 < x < 20)
            @icontract.require(lambda x: x % 7 == 0)
            def some_func(self, x: int) -> None:
                # The inputs from B.some_func need to satisfy either their own preconditions or
                # the preconditions of A.some_func ("require else").
                pass

            def __repr__(self) -> str:
                return "An instance of B"

        b = B()

        strategy = icontract_hypothesis.infer_strategy(b.some_func)

        self.assertEqual(
            "one_of("
            "fixed_dictionaries({"
            "'self': just(An instance of B),\n "
            "'x': integers(min_value=1, max_value=19)"
            ".filter(lambda x: x % 3 == 0)}), "
            "fixed_dictionaries({"
            "'self': just(An instance of B),\n "
            "'x': integers(min_value=1, max_value=19)"
            ".filter(lambda x: x % 7 == 0)}))",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(b.some_func)

    def test_composition(self) -> None:
        class A(icontract.DBC):
            @icontract.require(lambda x: x > 0)
            def __init__(self, x: int):
                self.x = x

            def __repr__(self) -> str:
                return "A(x={})".format(self.x)

        class B(icontract.DBC):
            @icontract.require(lambda y: y > 2020)
            def __init__(self, a: A, y: int):
                self.a = a
                self.y = y

            def __repr__(self) -> str:
                return "B(a={!r}, y={})".format(self.a, self.y)

        def some_func(b: B) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            "fixed_dictionaries("
            "{'b': "
            "fixed_dictionaries("
            "{'a': "
            "fixed_dictionaries("
            "{'x': integers(min_value=1)}).map(lambda d: A(**d)),\n"
            "  'y': integers(min_value=2021)}).map(lambda d: B(**d))})",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_abstract_class(self) -> None:
        class A(icontract.DBC):
            @abc.abstractmethod
            def do_something(self) -> None:
                pass

        # noinspection PyUnusedLocal
        class B(A):
            @icontract.require(lambda x: x > 0)
            def __init__(self, x: int):
                self.x = x

            def __repr__(self) -> str:
                return "B(x={})".format(self.x)

            def do_something(self) -> None:
                pass

        def some_func(a: A) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)

        # The strategies inferred for B do not reflect the preconditions of A.
        # This is by design as A is automatically registered with Hypothesis, so Hypothesis
        # will instantiate a B only at run time.

        self.assertEqual(
            "fixed_dictionaries("
            "{'a': fixed_dictionaries("
            "{'x': integers(min_value=1)})"
            ".map(lambda d: B(**d))})",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_enum(self) -> None:
        class A(enum.Enum):
            SOMETHING = 1
            ELSE = 2

        def some_func(a: A) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)

        self.assertTrue(
            re.match(
                r"fixed_dictionaries\({'a': sampled_from\([a-zA-Z_0-9.]+\.A\)}\)",
                str(strategy),
            ),
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_composition_in_data_class(self) -> None:
        class A(icontract.DBC):
            @icontract.require(lambda x: x > 0)
            def __init__(self, x: int):
                self.x = x

            def __repr__(self) -> str:
                return "A(x={})".format(self.x)

        @dataclasses.dataclass
        class B:
            a: A

        def some_func(b: B) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)

        # The strategies inferred for data classes do not reflect the preconditions of A.
        # This is by design as A is automatically registered with Hypothesis, so Hypothesis
        # will instantiate an A only at run time when creating B.
        self.assertEqual("fixed_dictionaries({'b': builds(B)})", str(strategy))

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_typed_dict(self) -> None:
        # TypedDict not available below Python version 3.8.
        if sys.version_info >= (3, 8):

            class A(icontract.DBC):
                @icontract.require(lambda x: x > 0)
                def __init__(self, x: int):
                    self.x = x

                def __repr__(self) -> str:
                    return "A(x={})".format(self.x)

            # noinspection PyTypedDict
            class B(TypedDict):
                a: A

            def some_func(b: B) -> None:
                pass

            strategy = icontract_hypothesis.infer_strategy(some_func)
            self.assertEqual(
                "fixed_dictionaries("
                "{'b': fixed_dictionaries("
                "{'a': fixed_dictionaries("
                "{'x': integers(min_value=1)})"
                ".map(lambda d: A(**d))}, optional={})})",
                str(strategy),
            )

            icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_list(self) -> None:
        class A(icontract.DBC):
            @icontract.require(lambda x: x > 0)
            def __init__(self, x: int):
                self.x = x

            def __repr__(self) -> str:
                return "A(x={})".format(self.x)

        def some_func(aa: List[A]) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            "fixed_dictionaries("
            "{'aa': lists("
            "fixed_dictionaries({'x': integers(min_value=1)})"
            ".map(lambda d: A(**d)))})",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_named_tuples(self) -> None:
        class A(icontract.DBC):
            @icontract.require(lambda x: x > 0)
            def __init__(self, x: int):
                self.x = x

            def __repr__(self) -> str:
                return "A(x={})".format(self.x)

        class B(NamedTuple):
            a: A

        def some_func(b: B) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)

        # The strategies inferred for named tuples do not reflect the preconditions of A.
        # This is by design as A is automatically registered with Hypothesis, so Hypothesis
        # will instantiate an A only at run time when creating B.
        self.assertEqual("fixed_dictionaries({'b': builds(B)})", str(strategy))

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_union(self) -> None:
        class A(icontract.DBC):
            @icontract.require(lambda x: x > 0)
            def __init__(self, x: int):
                self.x = x

            def __repr__(self) -> str:
                return "A(x={})".format(self.x)

        class B(icontract.DBC):
            @icontract.require(lambda x: x < 0)
            def __init__(self, x: int):
                self.x = x

            def __repr__(self) -> str:
                return "B(x={})".format(self.x)

        def some_func(a_or_b: Union[A, B]) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            "fixed_dictionaries("
            "{'a_or_b': one_of("
            "fixed_dictionaries({'x': integers(min_value=1)}).map(lambda d: A(**d)), "
            "fixed_dictionaries({'x': integers(max_value=-1)}).map(lambda d: B(**d)))})",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(some_func)


# noinspection PyUnusedLocal
class TestRepresentationOfCondition(unittest.TestCase):
    def test_that_a_single_line_condition_renders_correctly(self) -> None:
        # This test case was adapted from a solution for Advent of Code, day 8.
        class Operation(enum.Enum):
            NOP = "nop"
            ACC = "acc"
            JMP = "jmp"

        @dataclasses.dataclass
        class Instruction:
            operation: Operation
            argument: int

        @icontract.require(lambda instructions: len(instructions) < 10)
        def execute_instructions(instructions: List[Instruction]) -> Optional[int]:
            ...

        strategy = icontract_hypothesis.infer_strategy(execute_instructions)
        self.assertEqual(
            "fixed_dictionaries("
            "{'instructions': lists(builds(Instruction))"
            ".filter(lambda instructions: len(instructions) < 10)})",
            str(strategy),
        )

    def test_that_a_multi_line_condition_renders_correctly(self) -> None:
        # This test case was taken from a solution for Advent of Code, day 8.

        class Operation(enum.Enum):
            NOP = "nop"
            ACC = "acc"
            JMP = "jmp"

        @dataclasses.dataclass
        class Instruction:
            operation: Operation
            argument: int

        @icontract.require(
            lambda instructions: all(
                0 <= i + instruction.argument < len(instructions)
                for i, instruction in enumerate(instructions)
                if instruction.operation == Operation.JMP
            )
        )
        def execute_instructions(instructions: List[Instruction]) -> Optional[int]:
            ...

        strategy = icontract_hypothesis.infer_strategy(execute_instructions)
        self.assertEqual(
            """\
fixed_dictionaries({'instructions': lists(builds(Instruction)).filter(lambda instructions: all(
         0 <= i + instruction.argument < len(instructions)
         for i, instruction in enumerate(instructions)
         if instruction.operation == Operation.JMP
     ))})""",
            str(strategy),
        )

    def test_that_a_condition_as_function_renders_correctly(self) -> None:
        def some_condition(x: int) -> bool:
            ...

        @icontract.require(some_condition)
        def some_func(x: int) -> None:
            ...

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            "fixed_dictionaries({'x': integers().filter(some_condition)})",
            str(strategy),
        )


class TestSequence(unittest.TestCase):
    """Test that ``Sequence[T]`` is handled correctly.

    There is possibly a bug in Hypothesis 6.10.1 that causes ``binary()`` strategy on
    ``Sequence[int]``.
    """

    def test_sequence_int(self) -> None:
        # noinspection PyUnusedLocal
        def some_func(xs: Sequence[int]) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)

        # This might seem very surprising, but it is indeed the desired behavior,
        # see: https://github.com/HypothesisWorks/hypothesis/issues/2950
        self.assertEqual(
            "fixed_dictionaries({'xs': one_of(binary(), lists(integers()))})",
            str(strategy),
        )


class SomeGlobalClass(icontract.DBC):
    # We need this class so that we can try to infer the class of the unbound method based on
    # ``__qualname__`.

    def __init__(self) -> None:
        self.x = 1

    @icontract.require(lambda number: number > 0)
    @icontract.require(lambda self: self.x >= 0)
    def some_func(self, number: int) -> None:
        pass

    def __repr__(self) -> str:
        return f"An instance of {SomeGlobalClass.__name__}"


class SomeGlobalClassWithInheritance(SomeGlobalClass):
    @icontract.require(lambda another_number: another_number > 0)
    def another_func(self, another_number: int) -> None:
        pass

    def __repr__(self) -> str:
        return f"An instance of {SomeGlobalClassWithInheritance.__name__}"


class TestSelf(unittest.TestCase):
    """Test how ``self`` is generated in different settings."""

    def test_precondition_with_self_argument_and_another_argument(self) -> None:
        class A(icontract.DBC):
            def __init__(self) -> None:
                self.min_x = 0

            # noinspection PyShadowingNames
            @icontract.require(lambda self, x: self.min_x < x)
            def some_func(self, x: int) -> None:
                pass

            def __repr__(self) -> str:
                return "An instance of A"

        a = A()

        strategy = icontract_hypothesis.infer_strategy(a.some_func)
        self.assertEqual(
            "fixed_dictionaries({"
            "'self': just(An instance of A), 'x': integers()})"
            ".filter(lambda d: d['self'].min_x < d['x'])",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(a.some_func)

    def test_precondition_with_only_self_argument(self) -> None:
        class A(icontract.DBC):
            def __init__(self) -> None:
                self.x = 0

            # noinspection PyShadowingNames
            @icontract.require(lambda self: self.x >= 0)
            def some_func(self) -> None:
                pass

            def __repr__(self) -> str:
                return "An instance of A"

        a = A()

        strategy = icontract_hypothesis.infer_strategy(a.some_func)
        self.assertEqual(
            "fixed_dictionaries({"
            "'self': just(An instance of A)})"
            ".filter(lambda d: d['self'].x >= 0)",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(a.some_func)

    def test_unsatisfiable_precondition_with_self_argument(self) -> None:
        class A(icontract.DBC):
            def __init__(self) -> None:
                # This will make the pre-condition of ``some_func`` unsatisfiable.
                self.x = -1

            # noinspection PyShadowingNames
            @icontract.require(lambda number: number > 0)
            @icontract.require(lambda self: self.x >= 0)
            def some_func(self, number: int) -> None:
                pass

            def __repr__(self) -> str:
                return "An instance of A"

        a = A()

        strategy = icontract_hypothesis.infer_strategy(a.some_func)

        self.assertEqual(
            "fixed_dictionaries({"
            "'number': integers(min_value=1), "
            "'self': just(An instance of A)})"
            ".filter(lambda d: d['self'].x >= 0)",
            str(strategy),
        )

        error = None  # type: Optional[hypothesis.errors.FailedHealthCheck]
        try:
            icontract_hypothesis.test_with_inferred_strategy(a.some_func)
        except hypothesis.errors.FailedHealthCheck as err:
            error = err

        assert error is not None
        self.assertIsInstance(error, hypothesis.errors.FailedHealthCheck)

    def test_function_with_self_as_argument(self) -> None:
        class A:
            def __init__(self) -> None:
                self.x = 1

        # noinspection PyShadowingNames,PyUnusedLocal
        @icontract.require(lambda self: self.x > 10)
        def some_func(self: A) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)

        self.assertEqual(
            "fixed_dictionaries({'self': builds(A).filter(lambda self: self.x > 10)})",
            str(strategy),
        )

        error = None  # type: Optional[hypothesis.errors.Unsatisfiable]
        try:
            icontract_hypothesis.test_with_inferred_strategy(some_func)
        except hypothesis.errors.Unsatisfiable as err:
            error = err

        assert error is not None
        self.assertIsInstance(error, hypothesis.errors.Unsatisfiable)

    def test_precondition_on_self_on_unbound_instance_method_fails_on_nested_classes(
        self,
    ) -> None:
        class A(icontract.DBC):
            def __init__(self) -> None:
                self.x = 1

            # noinspection PyShadowingNames
            @icontract.require(lambda number: number > 0)
            @icontract.require(lambda self: self.x >= 0)
            def some_func(self, number: int) -> None:
                pass

            def __repr__(self) -> str:
                return f"An instance of {A.__name__}"

        # We can not infer the type of ``self`` as ``A`` is a nested class so we can not
        # "descend" to it from the top of the module based on ``__qualname__`` of ``some_func``.
        error = None  # type: Optional[TypeError]
        try:
            _ = icontract_hypothesis.infer_strategy(A.some_func)
        except TypeError as err:
            error = err

        assert error is not None

        got_error = re.sub(r"<function .*>", "<function ...>", str(error))

        self.assertEqual(
            "No search strategy could be inferred for the function: <function ...>; "
            "the following arguments are missing the type annotations: ['self'];\n\n"
            "sorted typed_args was ['number'], sorted parameter_set was ['number', 'self']",
            got_error,
        )

    def test_precondition_on_self_on_unbound_instance_method_with_localns(self) -> None:
        class A(icontract.DBC):
            def __init__(self) -> None:
                self.x = 1

            # We need to annotate ``self`` explicitly as we can not figure out the class from
            # an unbound method.
            # noinspection PyShadowingNames
            @icontract.require(lambda number: number > 0)
            @icontract.require(lambda self: self.x >= 0)
            def some_func(self: "A", number: int) -> None:
                pass

            def __repr__(self) -> str:
                return f"An instance of {A.__name__}"

        # We need to supply ``localns`` as ``self`` is annotated with a forward declaration and
        # ``A`` is a nested class.
        strategy = icontract_hypothesis.infer_strategy(A.some_func, localns={"A": A})

        self.assertEqual(
            "fixed_dictionaries({"
            "'number': integers(min_value=1),\n"
            " 'self': fixed_dictionaries({})"
            ".map(lambda d: A(**d))"
            ".filter(lambda self: self.x >= 0)})",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(A.some_func, localns={"A": A})

    def test_infer_self_if_class_is_not_nested(self) -> None:
        strategy = icontract_hypothesis.infer_strategy(SomeGlobalClass.some_func)

        self.assertEqual(
            "fixed_dictionaries({"
            "'number': integers(min_value=1),\n"
            " 'self': fixed_dictionaries({}).map(lambda d: SomeGlobalClass(**d))"
            ".filter(lambda self: self.x >= 0)})",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(SomeGlobalClass.some_func)

    def test_infer_self_if_class_inherits_and_unbound_instance_method(self) -> None:
        strategy = icontract_hypothesis.infer_strategy(
            SomeGlobalClassWithInheritance.another_func
        )

        self.assertEqual(
            "fixed_dictionaries({'another_number': integers(min_value=1),\n"
            " 'self': fixed_dictionaries({}).map(lambda d: SomeGlobalClassWithInheritance(**d))})",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(
            SomeGlobalClassWithInheritance.another_func
        )


if __name__ == "__main__":
    unittest.main()



"""Provide a valid module for testing pyicontract-hypothesis."""

from test_samples.pyicontract_hypothesis.sample_library import square_greater_than_zero

import icontract

SOME_FUNC_CALLS = 0


@icontract.require(lambda x: x > 0)
def some_func(x: int) -> None:
    global SOME_FUNC_CALLS
    SOME_FUNC_CALLS += 1


ANOTHER_FUNC_CALLS = 0


@icontract.require(lambda x: x > 0)
@icontract.require(
    lambda x: square_greater_than_zero(x), "dummy precondition to test filtering"
)
def another_func(x: int) -> None:
    global ANOTHER_FUNC_CALLS
    ANOTHER_FUNC_CALLS += 1


@icontract.require(lambda x, y: x < y, "Test multi-argument contracts")
def yet_another_func(x: int, y: int) -> None:
    pass

# pyicontract-hypothesis: disable
def expected_to_be_ignored(x: int) -> None:
    pass


# pyicontract-hypothesis: enable


# Class is ignored as well.
class A:
    @icontract.require(lambda x: x > 0)
    def __init__(self, x: int) -> None:
        self.x = x

    @icontract.require(lambda y: y > 0)
    @icontract.ensure(lambda result: result > 0)
    def some_method(self, y: int) -> int:
        return self.x + y
